/*->c.mpage */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <ctype.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"

#include "err.h"
#include "transform.h"
#include "etc.h"
#include "poll.h"


#include "constants.h"
#include "str.h"
#include "object.h"
#include "chain.h"
#include "undo.h"

#include "reslink.h"

#include "view.h"
#include "viewo.h"
#include "viewr.h"
#include "file.h"
#include "select.h"
#include "frameflow.h"

#include "txstr.h"
#include "fl.h"

#include "page.h"
#include "xpage.h"

#include "new.h"
#include "newx.h"
#include "modchap.h"
#include "chsect.h"

#include "mpage.h"



/* given master page object in src, find corresponding master page */
/* object in identical destination pages                           */


os_error * mapmaster(objectloc * src,objectloc * dest,
                                            int smaster,int * dmaster)
{
 os_error   * err;
 sectionstr * sectionp;
 filestr    * sfile;
 filestr    * dfile;
 int          n;

 err=NULL;

 *dmaster=0;


 if(smaster)
 {
  sfile=src->file;
  dfile=dest->file;

  sectionp=(sectionstr*)fob_find(src->section,sfile);

  n=objectindex(smaster,sectionp->dual,sfile);
  if(n>=0)
  {
   sectionp=(sectionstr*)fob_find(dest->section,dfile);
   *dmaster=objectfindindex(n,sectionp->dual,dfile);
  }
 }

 return(err);
}



static os_error * makepage(int msection,int mpage,
                           int dsection,int * dpage,int n,filestr * file)
{
 os_error     * err;
 objectloc      src;
 objectloc      dest;
 sectionstr   * sectionp;
 newpagestr     newpage;
 transformstr   tr;


 sectionp=(sectionstr*)fob_find(msection,file);
 newpage.data=sectionp->data;
 newpage.pagenumber=n;

 dest.file=file;
 dest.section=dsection;
 dest.page=0;
 dest.branch=0;
 dest.object=0;

 err=objectnew(&dest,OPAGE,&newpage);

 *dpage=dest.object;



 src.file=file;
 src.section=msection;
 src.page=mpage;
 src.branch=1;
 src.object=src.page;
 src.parent=src.page;
 src.after=0;
 src.chain=0;



 dest.file=file;
 dest.section=dsection;
 dest.page=*dpage;
 dest.branch=0;
 dest.object=0;
 dest.parent=*dpage;
 dest.after=0;
 dest.chain=0;
 dest.copymode=LOCFLOWDATA;

 tr_setidentity(&tr);

 err=objectcopy(&src,&dest,&tr);

 return(err);
}




/****************************************************************************/

os_error * scrapoldmaster(int dpage,filestr * file)
{
 pagestr    * pagep;
 int          nextdpage;

 if(dpage)
 {
  pagep=(pagestr*)fob_find(dpage,file);
  nextdpage=pagep->chain.next;

  scrappagecontents(dpage,file);
  scrappagecontents(nextdpage,file);
 }

 return(NULL);
}

static os_error * trasholdmasters(filestr * file)
{
 os_error   * err;
 int          dsection;
 int          dpage;
 sectionstr * sectionp;

 err=NULL;

 dsection=file->hdr.section[DOC];

 while(dsection)
 {
  sectionp=(sectionstr*)fob_find(dsection,file);
  dsection=sectionp->chain.next;

  dpage=sectionp->oldmaster;
  sectionp->oldmaster=0;

  scrapoldmaster(dpage,file);
 }

 return(err);
}




os_error * createoldmaster(int dsection,int msection,filestr * file)
{
 int          mpage;
 int          dpage;
 int          nextdpage;
 pagestr    * pagep;
 sectionstr * sectionp;


 sectionp=(sectionstr*)fob_find(msection,file);
 mpage=sectionp->chain.child;

 makepage(msection,mpage,dsection,&dpage,0,file);

 sectionp=(sectionstr*)fob_find(dsection,file);
 sectionp->oldmaster=dpage;

 relinkpage(dpage,dsection,file);

/* dprintf(0,"create old master %d",dpage); */

 pagep=(pagestr*)fob_find(mpage,file);
 mpage=pagep->chain.next;

 if(mpage)
 {
  makepage(msection,mpage,dsection,&nextdpage,1,file);
  insafter(nextdpage,dpage,0,0,file);
  relinkpage(nextdpage,dsection,file);
 }

 return(NULL);
}



static os_error * makeoldmasters(filestr * file)
{
 os_error   * err;
 int          msection;
 int          dsection;
 sectionstr * sectionp;

 err=NULL;

 dsection=file->hdr.section[DOC];
 msection=file->hdr.section[MST];

 while(dsection)
 {
  createoldmaster(dsection,msection,file);

  sectionp=(sectionstr*)fob_find(msection,file);
  msection=sectionp->chain.next;
  sectionp=(sectionstr*)fob_find(dsection,file);
  dsection=sectionp->chain.next;
 }

 return(err);
}




/*****************************************************************************/

/* find all the master frames on this page which are not on its master page */
/* as long as they don't have new data in, get rid of them                  */


static os_error * remmaster(int dpage,int dsection,
                            int mpage,int msection,filestr * file)
{
 os_error    * err;
 int           dchild;
 int           mchild;
 chainobject * link;
 pagestr     * pagep;
 int           next;
 int           master;
 objectbitstr  bits;
 int           chainn;


 err=NULL;

 for(chainn=0;chainn<2;chainn++)
 {
  pagep=(pagestr*)fob_find(dpage,file);

  if(chainn==0) dchild=pagep->chain.child;
  else          dchild=pagep->gchain.child;

  pagep=(pagestr*)fob_find(mpage,file);

  if(chainn==0) mchild=pagep->chain.child;
  else          mchild=pagep->gchain.child;

  while(dchild)
  {
   link=(chainobject*)fob_find(dchild,file);
   next=link->chain.next;

   err=objectmaster(dchild,file,&master,&bits);

   if(master && !objectinobjectchain(master,mchild,file))
   {
    objectlocal(dchild,file,DELETEMASTER);
   }

   dchild=next;
  }
 }

 return(err); 

 USE(msection);
 USE(dsection);
}












static os_error * unaddmaster(int dpage,int dsection,int ompage,
                              int mpage,int msection,filestr * file)
{
 os_error    * err;
 int           dchild;
 int           omchild;
 chainobject * link;
 pagestr     * pagep;
 int           next;
 int           master;
 objectbitstr  bits;
 int           chainn;
 int           message;


 err=NULL;


 for(chainn=0;chainn<2;chainn++)
 {
  pagep=(pagestr*)fob_find(dpage,file);

  if(chainn==0) dchild=pagep->chain.child;
  else          dchild=pagep->gchain.child;

  pagep=(pagestr*)fob_find(ompage,file);

  if(chainn==0) omchild=pagep->chain.child;
  else          omchild=pagep->gchain.child;

  while(dchild)
  {
   link=(chainobject*)fob_find(dchild,file);
   next=link->chain.next;

   err=objectmaster(dchild,file,&master,&bits);

   if(master && !bits.local)
   {
    message=OM_FINDMASTER;

    objectchainmessage(omchild,file,&message,&master);
    if(message)
    {
     objectdelete((objhdrstr*)fob_find(dchild,file),
                                         DELETESCRAP|DELETEUNLINK,file);
    }
   }
   dchild=next;
  }
 }

 return(err); 

 USE(mpage);
 USE(msection);
 USE(dsection);
}







/* scan master page, and create any objects on it, which are not on the     */
/* current page                                                             */

static os_error * addmaster(int dpage,int dsection,int ompage,
                            int mpage,int msection,filestr * file)
{
 os_error    * err;
 int           dchild;
 chainobject * link;
 pagestr     * pagep;
 int           next;
 int           master;
 int           mchild;
 int           omchild;
 int           lastinsert;
 int           message;
 objectloc     src;
 objectloc     dest;
 transformstr  tr;
 int           chainn;
 int           coff;

 err=NULL;

 for(chainn=0;chainn<2;chainn++)
 {
  pagep=(pagestr*)fob_find(dpage,file);

  if(chainn==0) 
  {
   dchild=pagep->chain.child;
   coff=0;
  }
  else
  {
   dchild=pagep->gchain.child;
   coff=PGLINECHAIN;
  }

  pagep=(pagestr*)fob_find(mpage,file);

  if(chainn==0)
  {
   mchild=pagep->chain.child;
  }
  else
  {
   mchild=pagep->gchain.child;
  }


  pagep=(pagestr*)fob_find(ompage,file);

  if(chainn==0)
  {
   omchild=pagep->chain.child;
  }
  else
  {
   omchild=pagep->gchain.child;
  }


  lastinsert=0;

  while(mchild)
  {
   link=(chainobject*)fob_find(mchild,file);
   next=link->chain.next;

   message=OM_FINDMASTER;
   master=mchild;

   objectchainmessage(omchild,file,&message,&master);
   if(message)
   {
    src.file=file;
    src.section=msection;
    src.page=mpage;
    src.branch=1;
    src.object=mchild;
    src.parent=mpage;
    src.chain=coff;


    dest.file=file;
    dest.section=dsection;
    dest.page=dpage;
    dest.branch=0;
    dest.object=0;
    dest.parent=dpage;
    dest.chain=coff;
    dest.copymode=LOCFLOWDATA;
    dest.after=lastinsert;

    tr_setidentity(&tr);

    err=objectcopy(&src,&dest,&tr);

    lastinsert=dest.object;
   }
   else
   {
    message=OM_FINDMASTER;
    master=mchild;

    objectchainmessage(dchild,file,&message,&master);
    if(!message) 
    {
     lastinsert=master;
    }
   }

   mchild=next;
  }
 }

 return(err);
}



static os_error * updatefrommaster(int dpage,int mode,filestr * file)
{
 os_error    * err;
 int           dchild;
 chainobject * link;
 pagestr     * pagep;
 int           next;
 int           master;
 objectbitstr  bits;
 int           chainn;

 err=NULL;

 for(chainn=0;chainn<2;chainn++)
 {
  pagep=(pagestr*)fob_find(dpage,file);

  if(chainn==0) dchild=pagep->chain.child;
  else          dchild=pagep->gchain.child;

  while(dchild)
  {
   link=(chainobject*)fob_find(dchild,file);
   next=link->chain.next;

   err=objectmaster(dchild,file,&master,&bits);

   if(master && !bits.local)
   {
    objectlocal(dchild,file,mode);
   }

   dchild=next;
  }
 }

 return(err); 
}






/*****************************************************************************/

typedef struct flushmstr
{
 int            cdsection;
 int            cmsection;
 sectiondatastr data;
 int            np;
} flushmstr;


static flushmstr cm;

typedef struct flushpstr
{
 int dpage;
} flushpstr;

static flushpstr cp;


static int       xmpage;
static int       xompage;
static int       xdble;
static int       xbase;
static int       xodble;
static int       xobase;
static int       xchangesize;



static os_error * mrelinkpage(filestr * file)
{
 os_error   * err;
 linkstr      link; 
 int          child;
 pagestr    * pagep;
 int          message;


 err=NULL;

 if(cp.dpage)
 {
  pagep=(pagestr*)fob_find(cp.dpage,file);
  child=pagep->chain.child;

/*  dprintf(0,"mrelink page=%d child=%d",cp.dpage,child); */

  link.dprev=pagep->chain.prev;
  link.dpage=cp.dpage;
  link.dnext=pagep->chain.next;
  link.dsection=cm.cdsection;

  link.mprev=link.mnext=0;

  if(undoforward(file))
  {
   link.mpage=xmpage;
   link.msection=cm.cmsection;

   if(xdble)
   {
    pagep=(pagestr*)fob_find(link.mpage,file);
    if(pagep->chain.next) link.mprev=link.mnext=pagep->chain.next;
    else                  link.mprev=link.mnext=pagep->chain.prev;
   }

   message=OM_RELINK;
  }
  else
  {
   link.mpage=xompage;
   link.msection=cm.cdsection;

   if(xodble)
   {
    pagep=(pagestr*)fob_find(link.mpage,file);
    if(pagep->chain.next) link.mprev=link.mnext=pagep->chain.next;
    else                  link.mprev=link.mnext=pagep->chain.prev;
   }

   message=OM_RELINKOM;
  }

  err=objectchainmessage(child,file,&message,&link);                 
 }

 return(err);
}





int findoldmaster(int master,filestr * file,int chain)
{
 int       message;
 pagestr * pagep;

 pagep=(pagestr*)fob_find(xompage,file);
 message=OM_FINDMASTER;

/*
 dprintf(2,"master=%d xompage=%d child=%d",master,xompage,pagep->chain.child);
*/

 objectchainmessage(chain?pagep->gchain.child:pagep->chain.child,
                                                       file,&message,&master);

/* dprintf(3,"master=%d",message?0:master); */

 return(message?0:master);
}




os_error * mpageundelete(mrefstr * mref,int coff,objectloc * src,
                         objectloc * dest,transformstr * tr,filestr * file)
{

 memset(src,0,sizeof(objectloc));
 memset(dest,0,sizeof(objectloc));

 src->file=file;
 src->section=cm.cdsection;
 src->page=xompage;
 src->object=findoldmaster(mref->master,file,coff>0);
 src->parent=cp.dpage;
 src->chain=coff;

 dest->file=file;
 dest->section=cm.cdsection;
 dest->page=cp.dpage;
 dest->parent=cp.dpage;
 dest->chain=coff;
 dest->copymode=LOCFLOWDATA|LOCOLDMASTER;
 dest->after=mref->after;

 tr_setidentity(tr);

 return(NULL);
}





os_error *  uuforcemaster(filestr * file)
{
/* dprintf(0,"uuforcemaster"); */

 undowritefn(file,URFORCEMASTER,NULL,0);
 return(updatefrommaster(cp.dpage,FORCEMASTERU,file));
}



os_error *  urforcemaster(filestr * file)
{
/* dprintf(0,"urforcemaster"); */

 undowritefn(file,UUFORCEMASTER,NULL,0);
 return(updatefrommaster(cp.dpage,FORCEMASTERR,file));
}






os_error *  uuaddmaster(filestr * file)
{
 unaddmaster(cp.dpage,cm.cdsection,xompage,xmpage,cm.cmsection,file);
 return(undowritefn(file,URADDMASTER,NULL,0));
}


os_error *  uraddmaster(filestr * file)
{
 addmaster(cp.dpage,cm.cdsection,xompage,xmpage,cm.cmsection,file);
 return(undowritefn(file,UUADDMASTER,NULL,0));
}








/* called at the end of each master section update */
/* tagged to UNFLUSHMS */

os_error * unflushms(filestr * file)
{


 if(xchangesize)
 {
  changepagesize(file,cm.cdsection);
  recalcmaxpagesize(file,DOC);
  fileresetviews(file,0);
 }

 return(undowritefn(file,UNFLUSHMF,&cm,sizeof(flushmstr)));
}




/* called at the start of each master section update */
/* tagged to UNFLUSHMF */

static os_error * unflushmfsub(filestr * file)
{
 sectiondatastr   data;
 int              np;
 sectionstr     * sectionp;

 sectionp=(sectionstr*)fob_find(cm.cdsection,file);
 data=sectionp->data;
 np=sectionp->np;

 xodble=sectionp->data.dble;
 xobase=getpagebase(sectionp);

 xchangesize=(data.width!=cm.data.width ||
              data.height!=cm.data.height);

 sectionp->data=cm.data;

 cm.data=data;
 cm.np=np;

 xdble=sectionp->data.dble;
 xbase=getpagebase(sectionp);

 return(undowritefn(file,UNFLUSHMS,NULL,0));
}


os_error * unflushmf(filestr * file)
{
 undoread(file,&cm,sizeof(flushmstr));

 return(unflushmfsub(file));
}





/* called at the end of each master page update */
/* tagged to UNFLUSHPS */

os_error * unflushps(filestr * file)
{
 mrelinkpage(file);

 frflend(file);
 vxend(file,VRBACK,cp.dpage,cm.cdsection);

 return(undowritefn(file,UNFLUSHPF,&cp,sizeof(flushpstr)));
}



/* called at the start of each master page update */
/* tagged to UNFLUSHPF */

os_error * unflushpf(filestr * file)
{
 pagestr    * pagep;
 sectionstr * sectionp;


 undoread(file,&cp,sizeof(flushpstr));

 vxstart(file);
 frflstart(file);

/* dprintf(0,"1"); */

 sectionp=(sectionstr*)fob_find(cm.cmsection,file);
 xmpage=sectionp->chain.child;

 if(xdble)
 {
  pagep=(pagestr*)fob_find(cp.dpage,file);

  if(!islhpb(pagep,xbase))
  {
   pagep=(pagestr*)fob_find(xmpage,file);
   xmpage=pagep->chain.next;
  }
 }

 sectionp=(sectionstr*)fob_find(cm.cdsection,file);
 xompage=sectionp->oldmaster;

 if(xodble)
 {
  pagep=(pagestr*)fob_find(cp.dpage,file);

  if(!islhpb(pagep,xobase))
  {
   pagep=(pagestr*)fob_find(xompage,file);
   xompage=pagep->chain.next;
  }
 }

 return(undowritefn(file,UNFLUSHPS,NULL,0));
}


/* need to change section data,
   but not pagenumber data */


static os_error * flushmasterpages(filestr * file)
{
 os_error   * err;
 int          msection;
 int          dsection;
 int          nmsection;
 int          ndsection;
 sectionstr * sectionp;
 pagestr    * pagep;
 int          nextpage;


 err=NULL;

 ndsection=dsection=file->hdr.section[DOC];
 nmsection=msection=file->hdr.section[MST];

/* dprintf(0,"flush master pages"); */

 while(nmsection)
 {
  cm.cdsection=ndsection;
  cm.cmsection=nmsection;

  sectionp=(sectionstr*)fob_find(ndsection,file);
  ndsection=sectionp->chain.next;
  cp.dpage=sectionp->chain.child;

  sectionp=(sectionstr*)fob_find(nmsection,file);
  nmsection=sectionp->chain.next;
  cm.data=sectionp->data;
  cm.np=sectionp->np;

  if(sectionp->changed)
  {
   createsectionmap(cm.cdsection,cm.cmsection,1,file);

   unflushmfsub(file);

  /* dprintf(0,"X0"); */

   while(cp.dpage)
   {
    pagep=(pagestr*)fob_find(cp.dpage,file);
    nextpage=pagep->chain.next;

    unflushpf(file);

    /* scan document page,                                          */
    /* rem/demaster any frames that do not appear in current master */

    remmaster(cp.dpage,cm.cdsection,xmpage,cm.cmsection,file);

   /* dprintf(0,"2"); */

    /* update any frames that do                                    */

    updatefrommaster(cp.dpage,FORCEMASTER,file);
    undowritefn(file,UUFORCEMASTER,NULL,0);
    updatefrommaster(cp.dpage,FORCEMASTER1,file);

   /* dprintf(0,"3"); */

    /* scan master and add to document any frames that do not exist */

    addmaster(cp.dpage,cm.cdsection,xompage,xmpage,cm.cmsection,file);
    undowritefn(file,UUADDMASTER,NULL,0);

   /* dprintf(0,"5"); */

    unflushps(file);

    cp.dpage=nextpage;
   }

   updatesectionmap(cm.cdsection,cm.cmsection,1,file);

   unflushms(file);
  }
 }

 return(err);
}


/*****************************************************************************/
/* three functions concerned with quiting master pages                       */


/* this undoes quiting master pages, open windows */
/* tagged to UNDOQUITMASTER */


os_error * undoquitscrap(filestr * file)
{
 int changed;
 int oldmaster;

 undoread(file,&changed,sizeof(int));

 while(1)
 {
  undoread(file,&changed,sizeof(int));
  if(changed<0) break;
  undoread(file,&oldmaster,sizeof(int));
  scrapoldmaster(oldmaster,file);
 }

 return(NULL);
}




os_error * undoquitmaster(filestr * file)
{
 os_error   * err;
 int          changed;
 int          oldmaster;
 sectionstr * sectionp;
 int          msection;
 int          prevmsection;
 int          dsection;
 int          prevdsection;


/* dprintf(3,"undoquitmaster"); */

 err=undowritefn(file,UNDOREQUITMASTER,NULL,0);

 undoread(file,&file->modded[MST],sizeof(int));

 msection=file->hdr.section[MST];
 dsection=file->hdr.section[DOC];
 while(msection)
 {
  prevmsection=msection;
  prevdsection=dsection;
  undoread(file,&changed,sizeof(int));
  undoread(file,&oldmaster,sizeof(int));

  sectionp=(sectionstr*)fob_find(msection,file);
  sectionp->changed=changed;
  msection=sectionp->chain.next;

  sectionp=(sectionstr*)fob_find(dsection,file);
  dsection=sectionp->chain.next;
  sectionp->oldmaster=oldmaster;

 /* dprintf(0,"read oldmaster %d changed=%d",oldmaster,changed); */

  if(!oldmaster) 
  {
   createoldmaster(prevdsection,prevmsection,file);
  }
 }

 err=openmasterview(file);

/* dprintf(3,"undoquitmasterx"); */

 return(err);
}





static os_error * savemasterstate(filestr * file)
{
 os_error   * err;
 int          changed;
 int          oldmaster;
 sectionstr * sectionp;
 int          msection;
 int          dsection;
 int          scrap;


 err=NULL;

 changed=file->modded[MST];
 file->modded[MST]=0;

 scrap=undoscrapcheck(file,(file->hdr.ns[MST]*2+2)*sizeof(int));

 if(!scrap) err=undowritefn(file,UNDOQUITMASTER,&changed,sizeof(int));

 msection=file->hdr.section[MST];
 dsection=file->hdr.section[DOC];
 while(msection)
 {
  sectionp=(sectionstr*)fob_find(msection,file);
  changed=sectionp->changed;
  sectionp->changed=0;
  msection=sectionp->chain.next;

  if(!scrap) undowrite(file,&changed,sizeof(int));

  sectionp=(sectionstr*)fob_find(dsection,file);
  dsection=sectionp->chain.next;
  oldmaster=sectionp->oldmaster;
  sectionp->oldmaster=0;

  if(!changed || scrap) 
  {
   scrapoldmaster(oldmaster,file);
   oldmaster=0;
  }

  if(!scrap) undowrite(file,&oldmaster,sizeof(int));
 }

 changed=-1;
 if(!scrap) undowrite(file,&changed,sizeof(int));

 return(err);
}



/* undo is closing master page windows */
/* tagged to UNDOREQUITMASTER          */



os_error * undorequitmaster(filestr * file)
{
 os_error * err;

/* dprintf(2,"requitmasteredit"); */

 err=undostart(file);
 if(!err)
 { 
  err=closemasterviews(file);
  if(!err) err=savemasterstate(file);
  undoend(file);
 }

 return(err);
}


/* external event closing master page windows */

os_error * quitmasteredit(filestr * file)
{
 os_error   * err;

/* dprintf(2,"quitmasteredit");  */

 err=undostart(file);
 if(!err)
 {
  event(EVENT_QUITMASTER,file,0);

  if(file->modded[MST])
  {
   clearselectionff(file,-1,-1,DOC);
   modfile(file,-1);
   flushmasterpages(file);
  }

  fileresetviews(file,0);
  file->hdr.maxpagewidth[DOC]=file->hdr.maxpagewidth[MST];
  file->hdr.maxpageheight[DOC]=file->hdr.maxpageheight[MST];

  savemasterstate(file);

  clearselectionff(file,-1,-1,MST);

  undoend(file);
 }

 return(err);
}


/*****************************************************************************/
/* three functions concerned with opening master pages                       */


/* this undoes quiting master pages, open windows */
/* tagged to UNDOREGOMASTER */

os_error * undoregomaster(filestr * file)
{
 os_error   * err;
 int          changed;
 sectionstr * sectionp;
 int          section;

/* dprintf(3,"undoregomaster");  */

 makeoldmasters(file);

 undoread(file,&file->modded[MST],sizeof(int));

 section=file->hdr.section[MST];
 while(section)
 {
  undoread(file,&changed,sizeof(int));
  sectionp=(sectionstr*)fob_find(section,file);
  sectionp->changed=changed;
  section=sectionp->chain.next;
 }

 err=undowritefn(file,UNDOGOMASTER,NULL,0);
 err=openmasterview(file);
/* dprintf(3,"undoregomasterx"); */

 return(err);
}



/* this undoes master pages, close windows */
/* tagged to UNDOGOMASTER */

os_error * undogomaster(filestr * file)
{
 os_error   * err;
 int          changed;
 sectionstr * sectionp;
 int          section;

/* dprintf(3,"undogomaster"); */

 trasholdmasters(file);

 err=closemasterviews(file);
 err=undowritefn(file,UNDOREGOMASTER,&file->modded[MST],sizeof(int));
 file->modded[MST]=0;

 section=file->hdr.section[MST];
 while(section)
 {
  sectionp=(sectionstr*)fob_find(section,file);
  changed=sectionp->changed;
  section=sectionp->chain.next;
  undowrite(file,&changed,sizeof(int));
  sectionp->changed=0;
 }

 return(err);
}



os_error * gomasteredit(filestr * file)
{
 os_error   * err;
 sectionstr * sectionp;
 int          section;

 err=undostart(file);
 if(!err)
 {
  makeoldmasters(file);

  event(EVENT_GOMASTER,file,0);

  undowritefn(file,UNDOGOMASTER,NULL,0);

  file->modded[MST]=0;

  section=file->hdr.section[MST];
  while(section)
  {
   sectionp=(sectionstr*)fob_find(section,file);
   section=sectionp->chain.next;
   sectionp->changed=0;
  }

 /* dprintf(2,"gomasteredit"); */

  undoend(file);
 }
 return(err);
}


