/*->c.transload */




#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 

#include "os.h"
#include "bbc.h"
#include "sprite.h"
#include "wimp.h"
#include "werr.h"
#include "transform.h"
#include "xprocess.h"


#include "wos.h"
#include "poll.h"
#include "fsx.h"
#include "bf.h"
#include "scrap.h"
#include "osreg.h"
#include "osthread.h"
#include "err.h"
#include "reslink.h"
#include "deb.h"
#include "bits.h"
#include "filetype.h"
#include "xext.h"


#include "etc.h"



#include "constants.h"
#include "str.h"
#include "pxstr.h"
#include "object.h"

#include "impex.h"

#include "transload.h"





// All of these functions return a pointer to the next token found
// in strToken. They return NULL when no more tokens are found. 
// Each call modifies strToken by substituting a NULL character
// for each delimiter that is encountered.

char * xstrtok(char *strToken,const char *strDelimit)
{
 static char * pointer;
 char * p;
 char * q;
 int    ch;
 int    c2;
 int    inquotes;

 if(strToken) pointer=strToken;
 else
 {
  if(!pointer) return(NULL);
  pointer=pointer+strlen(pointer)+1;
 }

 while(1)                    // strip any leading delimiters
 {
  ch=*pointer;
  if(!ch)
  {
   pointer=NULL;
   return(NULL);
  }
  
  q=(char*)strDelimit;
  while((c2=*q++)!=0)
  {
   if(ch==c2) break;
  }

  if(c2==0) break;
  pointer++;
 }

 p=pointer;
 inquotes=0;

 while(1)                  // search for end of token
 {
  ch=*p;
  if(!ch)
  {
   p=pointer;
   pointer=NULL;
   return(p);
  }

  if(ch=='"') inquotes^=1;
  if(!inquotes)
  {
   q=(char*)strDelimit;
   while((c2=*q++)!=0)
   {
    if(ch==c2) break;
   }

   if(c2!=0)
   {
    *p=0;
    break;
   }
  }
  p++;
 }

 return(pointer);
}




// given source and destination extension, locate a filter

static os_error * translatereadx(char * sext,char * dext,trans_info * info,int flags)
{
 os_error * err;
 oskeyh     key;
 char       name[256];
 int        namesize;
 int        valuesize;
 int        valuetype;
 int        eof;
 int        found=0;
 char     * tok;
 char     * pointer;
 char       buffer[FILENAMELEN];
 char     * s;
 char     * d;
 int        localflags;


 err=regopenkey("HKEY_LOCAL_MACHINE\\Software\\DavidPilling\\Filters",&key,'r');
 if(err)
 {
  err=geterror(ENOFL);
 }
 else
 {
  found=0;
  
  while(1)
  {
   namesize=sizeof(name);
   valuesize=sizeof(buffer);

   err=regenumvalue(key,name,&namesize,&valuetype,buffer,&valuesize,&eof,&info->context_ptr);
   if(err) break;

//   cprintf("trans %s eof %d",eof?"":name,eof);
   
   if(eof)
   {
    err=geterror(ENOTRANS);
    break;
   }

   pointer=buffer;
			localflags=0;

   while((tok=xstrtok(pointer," \t"))!=NULL)
   {
    pointer=NULL;
//    cprintf("tok %s",tok);

				if(strncmp(tok,"-C",2)==NULL)
				{
					localflags|=TRANS_CONTENT;
				}
				else
    if(strncmp(tok,"-T",2)==NULL)
    {
     s=xstrtok(NULL," \t"); // got to always strtok twice after -T
     d=xstrtok(NULL," \t");
      
     if(!found)
     {
     /* get source filetype */

//						cprintf("sext %s s %s",s,sext);

      found=(cstrcmp(s,sext)==0);

     /* get destination filetype */

      if(dext!=NULL && cstrcmp(d,dext)) found=0;

//						cprintf("dext %s d %s",d,dext);

      if(found)
      {
       getfiletype(&info->dtype,d);
      }
     }
    }
    else
    if(strncmp(tok,"-R",2)==NULL)
    {
     s=xstrtok(NULL," \t"); // got to always strtok after -R

     if(found)
     {
						if((localflags ^ flags)==0)
						{
       strcpy(info->translator,s);
       break;
						}
     }
    }

   }
   if(found && ((localflags ^ flags)==0)) break;
  }
  err=regclosekey(&key,err);
 }

//	cprintf("found %d err %x trans %s",found,err,found?info->translator:"");

 return(err);


}


static os_error * runtranslator(char * sname,char * dname,trans_info * info,int flags)
{
 os_error * err;
 char       buffer[3*FILENAMELEN];
 int        ftype;
 int        exitcode=0;

 /* start the translator if it exists */

 err=fs_exists(info->translator,&ftype);
	if(!err)
	{
  if(ftype!=1) err=geterror(EFLBAD);
  else
 {
   err=xsprintf(sizeof(buffer),NULL,buffer,"%s %s \"%s\" \"%s\"",info->translator,(flags & TRANS_CONTENT)?"-C":"",sname,dname);

//  cprintf("tx %s",buffer);

   if(!err) err=runexe(buffer,&exitcode);

//  cprintf("exitcode %x filenamelen %d",exitcode,FILENAMELEN);

   if(!err) err=geterroratom(exitcode);
  }
	}

 return(err);
}


// locate a translator between known source and dest types
// used to answer the question can we translate source to something
// input dtype of -1 signifies any destination type

os_error * translatereadflags(int stype,int dtype,trans_info * info,int flags)
{
 os_error * err;
 char       sext[FSMAXEXTLEN];
 char       dext[FSMAXEXTLEN];
 int        i;
 int        j;
 int        hit=0;

 err=NULL;

 j=0;
 while(1)
 {
  err=getextensionx(stype,sext,j,&hit);
  if(err) break;
  if(hit==0)
  {
   err=geterror(ENOTRANS);
   break;
  }

  i=0;
  while(1)
  {
   err=NULL;
   if(dtype>=0) err=getextensionx(dtype,dext,i,&hit);
   if(err) break;
   if(hit==0)
   {
    err=geterror(ENOTRANS);
    break;
   }

   err=translatereadx(sext,(dtype<0)?NULL:dext,info,flags);

//   cprintf("readx %s %s err %x",sext,(dtype<0)?"":dext,err);

   if(!err || dtype<0) break;

   i++;
  }

  if(!err) break;

  j++;
 }

 return(err);
}


os_error * translateread(int stype,int dtype,trans_info * info)
{
	return(translatereadflags(stype,dtype,info,0));
}


// locate a translator between unknown source and known dest types
// used to answer the question can we translate source to something
// input dtype of -1 signifies any destination type


static os_error * translateread2(char * sext,int dtype,trans_info * info,int flags)
{
	os_error * err;
	int        i;
 char       dext[FSMAXEXTLEN];
	int        hit;

 err=NULL;


	i=0;
	while(1)
	{
  getextensionx(dtype,dext,i,&hit);

  if(hit==0)
  {
   err=geterror(ENOTRANS);
   break;
  }

  info->context_ptr=0;
  err=translatereadx(sext,dext,info,flags);
		if(!err) break;
		i++;
	}

	return(err);
}






// function to test if we can convert sname to destination type
// extract source type from sourcename extension
// signal success by error returned

os_error * translatetest(char * sname,int dtype)
{
 os_error * err;
 char       dext[FSMAXEXTLEN];
 char     * sext;
 trans_info info;
 int        i;
 int        hit=0;
 int        stype;

 err=NULL;

//	cprintf("translate test %s",sname);


 info.context_ptr=0;

 stype=fs_filetypefromextensionx(sname);  // do we know about this extension
 if(stype>=0)                             // if so may be duplicate versions
 {																															 								// one file type may have many extensions
//		cprintf("know about extension %x",stype);

  err=translateread(stype,dtype,&info);
 }
 else
 {
  sext=fs_extension(sname);
  if(!sext) err=geterror(ENOTRANS);
  else
  {
   i=0;
   while(1)
   {
    err=NULL;

    if(dtype>=0) err=getextensionx(dtype,dext,i,&hit);
  		if(err) break;
    if(hit==0)
    {
     err=geterror(ENOTRANS);
     break;
    }

//    cprintf("extensions %s %s",sext,(dtype<0)?"":dext);

    err=translatereadx(sext,(dtype<0)?NULL:dext,&info,0);
    if(!err || dtype<0) break;

    i++;
   }
  }
 }

 return(err);
}




os_error * translatetest2(int stype,int dtype,int flags)
{
 os_error * err;
 trans_info info;

 err=NULL;

 info.context_ptr=0;

 err=translatereadflags(stype,dtype,&info,flags);

 cprintf("translate test2 returns %d flags %x s %x d %x",(err!=NULL),flags,stype,dtype);

 return(err);
}









// translate from one known filetype to another

os_error * translatefile(char * sname,int stype,char * dname,int dtype)
{
 os_error  * err;
 trans_info  info;
 char      * sext;

 err=NULL;

// cprintf("translate read");

 info.context_ptr=0;
 if(stype>=0) err=translateread(stype,dtype,&info);
 else
 {
  sext=fs_extension(sname);
  if(sext) err=translateread2(sext,dtype,&info,0);
  else     err=geterror(ENOTRANS);
 }

 if(!err) err=runtranslator(sname,dname,&info,0);

 return(err);
}



// int transload(string & from,int ftype,string & to,int ttype);
// return 0 on success

int transloadx(int fp)
{
 os_error * err;

 err=translatefile(stringptr(stack[fp]),stack[fp+1],stringptr(stack[fp+2]),stack[fp+3]);
 zreport(err);
 return(0);
}



// subroutine to scan through possible extensions for destination type





// input source name and destination name
// output converted file and destination type
// scan through picture types and finally DDL

os_error * transloadflags(char * sname,char * dname,int * dtype,int flags)
{
 os_error   * err;
 trans_info   info;
 char       * sext;
 int          stype;
	int          i;

 err=NULL;

	cprintf("transload %s %s",sname,dname);

 info.context_ptr=0;

 stype=fs_filetypefromextensionx(sname);  // do we know about this extension
 if(stype>=0)                            // if so may be duplicate versions
 {
  cprintf("translate read stype %x",stype);
  err=translatereadflags(stype,-1,&info,flags);
  if(!err) *dtype=info.dtype;
 }
 else
 {
  sext=fs_extension(sname);
  if(sext)
  {
   i=0;
 		while(1)
 		{
 			*dtype=impexftypeindex(i,IMPXPICT|IMPXLOAD);
 			if(*dtype<0)
    {
 				err=geterror(ENOTRANS);
     break;
 			}
	 		if(*dtype!=0)
	 		{
//		   cprintf("dtype %x sext %s",*dtype,sext);
	   	err=translateread2(sext,*dtype,&info,flags);
		 	 if(!err) break;
		 	}

		 	i++;
		 }

   if(err && err->errnum==ENOTRANS)
   {
//    cprintf("Dtype2");
    err=translateread2(sext,*dtype=DTYPE2,&info,flags);
//    if(err) cprintf("transerr");
   }
  }
  else
  {
   err=geterror(ENOTRANS);
  }
 }


//	cprintf("dtype %x",*dtype);

//	cprintf("run trans %s %s",sname,dname);

 if(!err) err=runtranslator(sname,dname,&info,flags);

 return(err);
}



os_error * transload(char * sname,char * dname,int * dtype)
{
	return(transloadflags(sname,dname,dtype,0));
}
