/*->c.xtrace */



#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <locale.h>
#include <math.h>


#include "os.h"
#include "wimp.h"
#include "sprite.h"
#include "wimpt.h"
#include "bbc.h"
#include "akbd.h"
#include "werr.h"
#include "flex.h"


#include "filetype.h"
#include "wos.h"
#include "colour.h"
#include "transform.h"
#include "xprocess.h"
#include "etc.h"
#include "bf.h"
#include "hourglass.h"
#include "err.h"
#include "deb.h"
#include "sprx.h"
#include "imlib.h"
#include "bmpim.h"
#include "xmath.h"
#include "poll.h"
#include "dbhi.h"

#include "constants.h"
#include "str.h"
#include "reslink.h"

#include "xxcolour.h"
#include "file.h"
#include "object.h"
#include "viewr.h"
#include "colsep.h"

#include "geometry.h"
#include "xredraw.h"
#include "curve.h"
#include "curvelo.h"

#include "pxstr.h"
#include "pict.h"


#include "splib.h"

#include "impex.h"

#include "fit.h"
#include "xtrace.h"

#include "xtracex.h"



/*****************************************************************************/


static char * spseenmask=NULL;


static os_error * initseenmask(xtracestr * xstr)
{
 int seensize;
 int * i;
 os_error * err;

 seensize=(xstr->xpix*xstr->ypix);
 seensize=(seensize>>3)+((seensize & 7)>0);
 seensize=(seensize>>2)+((seensize & 3)>0);

 err=flex_alloc((flex_ptr)&spseenmask,seensize*4);

 if(!err)
 {
  i=(int *)spseenmask;
  while(seensize--) *i++=0;
 }

 return(err);
}


static void trashseenmask(xtracestr * xstr)
{
 flex_free((flex_ptr)&spseenmask);
 USE(xstr);
}


static int testseenmask(xtracestr * xstr,int x,int y)
{
 int bit=x+y*xstr->xpix;
 int byte=*(spseenmask+(bit>>3));
 bit=bit & 0x7;
 return((byte>>bit) & 0x1);
}




/* return value of pixel */

static int pixelxy(xtracestr * xstr,int x,int y)
{
 int  * data;
 int    bits;
 int    word;

 if(x<0 || x==xstr->xpix || y<0 || y==xstr->ypix) return(-1);

 data=(int*)(xstr->datap+xstr->offset+xstr->spdata);

 if(!xstr->topdown) y=(xstr->ypix-1)-y;

 data+=y*xstr->spwidth; /* pointed at line */

 bits=x<<xstr->spbpp;
 bits=bits+xstr->leftwastage;
 word=bits>>5;         // div 32 to get words from bits
 bits=bits & 0x1F;     // mask with 31

 if(xstr->rightleft)
 {
  bits=(bits & 0x18)+(7-(bits & 0x7));
 }

 data+=word;

 return(((*data)>>bits) & xstr->spbitmask);
}



static void setseenmask(xtracestr * xstr,int x,int y)
{
 int     bit=x+y*xstr->xpix;
 char *  byte=(spseenmask+(bit>>3));
 bit=bit & 0x7;
 *byte|=(0x1 << bit);
}


static int pixelmxy(xtracestr * xstr,int x,int y,int target)
{            
 if(x<0 || x>=xstr->xpix || y<0 || y>=xstr->ypix) return(target);
 if(testseenmask(xstr,x,y)) return(-1);
 return(pixelxy(xstr,x,y));
}


/*****************************************************************************/


#define INITPSIZE 2560
#define STEPPSIZE 2560


static os_error * initpointstore(xtracestr * xstr)
{
 os_error * err;

 err=flex_alloc((flex_ptr)&xstr->fit.pointstore,INITPSIZE);
 if(!err)
 {
  xstr->fit.npoints=0;
  xstr->fit.maxpoints=INITPSIZE/(sizeof(coordstr))-2;
 }

 return(err);
}


static void clearpointstore(xtracestr * xstr)
{
 xstr->fit.npoints=0;
}



static void trashpointstore(xtracestr * xstr)
{
 flex_free((flex_ptr)&xstr->fit.pointstore);
}



static os_error * addpoint(xtracestr * xstr,int x,int y)
{
 os_error * err;

 err=NULL;

 xstr->fit.pointstore[xstr->fit.npoints].x=x;
 xstr->fit.pointstore[xstr->fit.npoints].y=y;
 xstr->fit.npoints++;

 if(xstr->fit.npoints>=xstr->fit.maxpoints)
 {
  xstr->fit.maxpoints+=STEPPSIZE/sizeof(coordstr);
  err=flex_extend((flex_ptr)&xstr->fit.pointstore,
                             xstr->fit.maxpoints*sizeof(coordstr));
 } 

 return(err);
}



/*****************************************************************************/


typedef struct xpointstr
{
 coordstr c;
 int      n;
} xpointstr;



static int pointcmp(const void * p1,const void * p2)
{
 if((((xpointstr*)p1)->c.y)==(((xpointstr*)p2)->c.y))
   return((((xpointstr*)p1)->c.x)-(((xpointstr*)p2)->c.x));
 else
   return((((xpointstr*)p1)->c.y)-(((xpointstr*)p2)->c.y));
}




static void fillareax(xtracestr * xstr)
{
 int        i;
 int        xs;
 int        xf;
 int        ys;
 int        x;
 int        y;
 int        npoints;
 int        spddx;
 int        spddy;
 int        n;
 int        s1;
 int        s2;
 xpointstr * p;
 int        outside;
 int        next;
 int        prev;
 int        nexty;
 int        prevy;



 /* sort points by y then x */

 npoints=xstr->fit.npoints;
 spddx=xstr->fit.spddx;
 spddy=xstr->fit.spddy;

 flex_alloc((flex_ptr)&p,npoints*sizeof(xpointstr));
 for(i=0;i<npoints;i++) 
 {
  p[i].c=xstr->fit.pointstore[i];
  p[i].n=i;
 }

 qsort(p,npoints,sizeof(xpointstr),pointcmp);

 i=0;

 while(i<npoints)
 {
  y=p[i].c.y;
  ys=y/spddy;
  outside=1;

  if(ys>=0 && ys<xstr->ypix)
  {
   while(i<npoints && p[i].c.y==y)
   {
    xs=p[i].c.x/spddx;
    if(xs>=0 && xs<xstr->xpix) setseenmask(xstr,xs,ys);

    n=p[i].n;
    next=n+1;
    if(next>=npoints) next=0;
    prev=n-1;
    if(prev<0) prev=npoints-1;
    prevy=xstr->fit.pointstore[prev].y;
    nexty=xstr->fit.pointstore[next].y;
    s1=nexty-y;
    s2=prevy-y;

    if(s1 && s2)
    {
     if(s1!=s2)
     {
      outside^=1;
     }
    }
    else
    if(s2)
    {
     while(1)
     {
      next++;
      if(next>=npoints) n=0;
      if((s1=xstr->fit.pointstore[next].y-y)!=0) break;
     }
     if(s1!=s2)
     {
      outside^=1;
     }
    }


    if(i<(npoints-1) && p[i+1].c.y==y)
    {
     /* span */

     xf=p[i+1].c.x/spddx;

     if(xs<(xf-1))
     {
      if(!outside)
      {
       if(xs<0) xs=0;
       if(xf<0) xf=0;
       if(xs>=xstr->xpix) xs=xstr->xpix-1;
       if(xf>=xstr->xpix) xf=xstr->xpix-1;

       for(x=xs;x<xf;x++)
       {
        setseenmask(xstr,x,ys);     
       }
      }
     }
    }
    i++;
   }
  }
  else
  {
   while(i<npoints && p[i].c.y==y) i++;
  }
 }

 flex_free((flex_ptr)&p);

}

/*****************************************************************************/


static os_error * getoutline(xtracestr * xstr,int x,int y,int target)
{
 int        i;
 int        dir;
 int        sx;
 int        sy;
 int        dx;
 int        dy;
 os_error * err;

 dir=0;
 sx=x;
 sy=y;
 dx=dy=0;    /* compiler */
 err=NULL;

 do
 {
  i=dir;
  do
  {
   switch(i & 0x7)
   {
    case 0:
           dx=-1;
           dy=1;
           break;

    case 1:
           dx=0;
           dy=1;
           break;

    case 2:
           dx=1;
           dy=1;
           break;

    case 3:
           dx=1;
           dy=0;
           break;

    case 4:
           dx=1;
           dy=-1;
           break;

    case 5:
           dx=0;
           dy=-1;
           break;

    case 6:
           dx=-1;
           dy=-1;
           break;

    case 7:
           dx=-1;
           dy=0;
           break;
   }

   if(pixelmxy(xstr,x+dx,y+dy,target)!=target) {i=i & 0x7;break;}
   i++;

  }while(!(i>(dir+7)));

  if(i>(dir+7))
  {
   dx=dy=0;
  }

  x+=dx;
  y+=dy;

  err=addpoint(xstr,x*xstr->fit.spddx,y*xstr->fit.spddy);
  if(err) break;

  dir=(i+5) & 0x7;

 }while(!(x==sx && y==sy));

 return(err);
}




static os_error * fitoutline(xtracestr * xstr,int target)
{
 os_error * err;
 int        x;
 int        y;
 int        pixel;

 err=NULL;

 if(target<0) target=pixelmxy(xstr,0,0,0);

 for(y=xstr->ypix-1;y>=0;y--)
 {
  for(x=0;x<xstr->xpix;x++)
  {
   if(!testseenmask(xstr,x,y))
   {
    pixel=pixelmxy(xstr,x,y,target);

    if(pixel!=target)
    {
     clearpointstore(xstr);
     err=addpoint(xstr,x*xstr->fit.spddx,y*xstr->fit.spddy);
     if(err) break;

     err=getoutline(xstr,x,y,target);
     if(err) break;

     if(xstr->fit.npoints>4)
     {
     /* dprintf(0,"fit in"); */
      err=fitcurve(&xstr->fit,0,xstr->fit.npoints);
     /* dprintf(0,"fit out"); */
      if(!err)
      {
       /* do something with points */

       vecclose(&xstr->fit);
      }
     }

     if(!err)
     {
    /* dprintf(0,"fill in"); */
      fillareax(xstr);
    /* dprintf(0,"fill out"); */
     /* x=xstr->pix.spx;
      y=0; */
     }
    }
   } 

//   if(bbc_inkey(0)==27) err=geterror(ESCAPERR);
   if(akbd_pollesc()) err=geterror(ESCAPERR);
   if(err) break;
  }
  if(err) break;
 }

 return(err);
}




static os_error * dofit(xtracestr * xstr,int target)
{
 os_error * err;

 err=initseenmask(xstr);

 if(!err)
 {
  err=initpointstore(xstr);

  if(!err)
  {
   if(!err)
   {
    if(!err)
    {
     if(!err) 
     {
      err=fitoutline(xstr,target);
     }
    }
   }
   trashpointstore(xstr);
  }
  trashseenmask(xstr);
 }

 return(err);
}


#define DRAWWORDS 16

static char dpitab[]={0,180,90,60,45,22};


os_error * getpixinfo(newsprite_header * spa,pixinfostr * px)
{
 sprite_id       sid;
 sprite_info     sinfo;

 sid.s.addr=(sprite_area *)spa;
 sid.tag=sprite_id_addr;
 sprite_readsize((sprite_area*)sid.s.addr,&sid,&sinfo);

 px->spx=sinfo.width;
 px->spy=sinfo.height;
 px->spmode=sinfo.mode;
 px->sphasmask=sinfo.mask;

 getdeltas(px->spmode,&px->spdx,&px->spdy);
 getbpp(px->spmode,&px->spbpp,&px->spbitmask);

// cprintf("spbpp %d mode %x",px->spbpp,px->spmode);
 
 px->spbpp=ln2(px->spbpp);



 px->sposx=px->spx*px->spdx;
 px->sposy=px->spy*px->spdy;

 px->spdrx=px->sposx*256;                         /* draw units */
 px->spdry=px->sposy*256;

 px->spddx=px->spdx*256;
 px->spddy=px->spdy*256;

 px->spleft=spa->lbit;
 px->spright=spa->rbit;

 px->spwidth=spa->width+1;
 px->spheight=spa->height+1;
 px->spimage=spa->image;

 px->palentries=(spa->image-sizeof(sprite_header))/8;

 if(spa->mask==spa->image) px->spmask=0;
 else                      px->spmask=spa->mask;

 px->sptype=((newsprite_header*)spa)->selector.type.type;

 if(px->sptype)
 {
  px->spleft=0;
  px->spmaskbpp=0;
  px->spmaskbitmask=1;
  px->spmaskwidth=(px->spx+31)/32;
  px->xdpi=((newsprite_header*)spa)->selector.type.xdpi;
  px->ydpi=((newsprite_header*)spa)->selector.type.ydpi;
 }
 else
 {
  px->spmaskbpp=px->spbpp;
  px->spmaskbitmask=px->spbitmask;
  px->spmaskwidth=px->spwidth;
  px->xdpi=dpitab[px->spdx];
  px->ydpi=dpitab[px->spdy];
 }


 return(NULL);
}



/* create picture bounding curve */

/* (1) is this a vector file - yes create a mask and render to it */
/* (2) is this a bit map >8bpp forget it                          */
/* (3) has it got a mask - yes trace the mask                     */
/* (4) get background, and ignore this colour, trace any non-background */


os_error * xtrace(int pict,filestr * file)
{
 os_error       * err;
 pictboxstr     * pictp;
 xtracestr        xstr;
 curvestr         c;
 int              xpicture;
 picturestr     * picturep;
 int              xpix;
 int              ypix;
 aboxstr          dbox;
 transformstr     tr;
 transformstr     tra;
 transformstr     trb;
 processstr       pr;
 int              ctrans;
 int              masked;
 viewredrawstr    v;
// wimp_palettestr  wpal;
 imagestr         im;
 imagescreenstr   imhandle;
 pixinfostr       pix;
 int              target;


 err=NULL;
 ctrans=0;
 masked=0;
 hourglasson();

 pictp=(pictboxstr*)fob_find(pict,file);
 if(pictp)
 {
  c=pictp->pcurve;
  xpicture=pictp->picture;
  vecopen(&xstr.fit);

  if(pictp->ftype==SPRITE)
  {
   err=fob_lock((fobjptr)&xstr.datap,xpicture,file);
   if(!err)
   {
    xstr.offset=sizeof(picturestr)-sizeof(int)+DRAWWORDS*4;

    getpixinfo((newsprite_header*)(xstr.datap+xstr.offset),&pix);
    masked=pix.sphasmask;
    
    xstr.fit.spddx=pix.spddx;
    xstr.fit.spddy=pix.spddy;
    xstr.xpix=pix.spx;
    xstr.ypix=pix.spy;
    xstr.leftwastage=pix.spleft;
    xstr.topdown=xstr.rightleft=0;

    if(pix.spbpp>8) err=geterror(EPX7);
    if(!err)
    {
     if(pix.sphasmask)
     {
      xstr.spdata=pix.spmask;
      xstr.spbitmask=pix.spmaskbitmask;
      xstr.spbpp=pix.spmaskbpp;
      xstr.spwidth=pix.spmaskwidth;
      target=0;
     }
     else
     {
      xstr.spdata=pix.spimage;
      xstr.spbitmask=pix.spbitmask;
      xstr.spbpp=pix.spbpp;
      xstr.spwidth=pix.spwidth;
      target=-1;
     }   
     
     err=dofit(&xstr,target);
     if(xstr.fit.pathsize==0) err=geterror(EPX0);
    }

    fob_unlock((fobjptr)&xstr.datap,file);

    vecend(&xstr.fit);
   }
  }
  else
  if(pictp->ftype==BMP)
  {
   err=fob_lock((fobjptr)&xstr.datap,xpicture,file);
   if(!err)
   {
    xstr.offset=sizeof(picturestr)-sizeof(int)+2;

    
    xstr.fit.spddx=pix.spddx;
    xstr.fit.spddy=pix.spddy;
    xstr.xpix=pix.spx;
    xstr.ypix=pix.spy;
    xstr.leftwastage=0;
    xstr.topdown=xstr.rightleft=0;

    if(pix.spbpp>8) err=geterror(EPX7);
    if(!err)
    {
     xstr.spdata=pix.spimage;
     xstr.spbitmask=pix.spbitmask;
     xstr.spbpp=pix.spbpp;
     xstr.spwidth=pix.spwidth;
     
     err=dofit(&xstr,0);
     if(xstr.fit.pathsize==0) err=geterror(EPX0);
    }

    fob_unlock((fobjptr)&xstr.datap,file);

    vecend(&xstr.fit);
   }
  }
  else
  {
   initviewredraw(&v);
   memset(&pr,0,sizeof(processstr));

   v.file=file;        /* !!!! more fields need setting up here */


   masked=1;

   picturep=(picturestr*)fob_find(xpicture,file);
   xpix=(picturep->abox.x1-picturep->abox.x0)/(256);     // 180 dpi image
   ypix=(picturep->abox.y1-picturep->abox.y0)/(256);
   dbox=picturep->abox;


   im.xdpi=180;
   im.ydpi=180;
   im.ipal.word[0]=0xFFFFFF00;
   im.ipal.word[1]=0x0;
   im.ipal.ncolours=2;
   im.ipal.palclass=CL_RGB;

   err=imnewscreen(xpix,ypix,1,&im,&imhandle);
   if(!err)
   {
    imsetscreen(&imhandle,1);
    if(!err)
    {
     pr.mode=PROCESSBLACK;
     pr.isprinter=0;
//     pr.vdupal=&wpal;
//     pr.deltax=1;                                        // 180 dpi image
//     pr.deltay=1;
//     pr.lndeltax=0;
//     pr.lndeltay=0;
     pr.xdpi=180;
     pr.ydpi=180;
     pr.ln2bpp=0;
     pr.tosprite=1;
     pr.scale.x=0x10000;
     pr.scale.y=0x10000;
     pr.fontbits=0;
     pr.jpegbits=0;
     pr.spritebits=0;
     pr.useproxies=0;

     pr.handle2=NULL; // view;
     pr.cmyk2rgb=(processcmyk2rgbfn)xcmyk2rgb3232;
     pr.cmyk2bgr=(processcmyk2rgbfn)xcmyk2bgr3232;



     tr_setscale(&tra,dbox.x1-dbox.x0,scale(xpix,180*256,cvdu.xdpi),
                      dbox.y1-dbox.y0,scale(ypix,180*256,cvdu.ydpi));
     tr_setshift(&trb,-dbox.x0,-dbox.y1);

     tr_multiply(&tr,&trb,&tra);

     picturep=(picturestr*)fob_find(xpicture,file);
     err=render(&dbox,&tr,&pr,picturep,&v);

     imsetscreen(&imhandle,0);

     xstr.fit.spddx=256;
     xstr.fit.spddy=256;
     xstr.xpix=xpix;
     xstr.ypix=ypix;
     xstr.leftwastage=0;
     xstr.spdata=0;
     xstr.offset=0;
     xstr.spbitmask=1;
     xstr.spbpp=0;
     xstr.spwidth=im.wwidth/4;
     xstr.datap=(char*)im.data;
     xstr.topdown=0;
     xstr.rightleft=1;

//     savebmp("c:\\Temp\\xbmp.bmp",&im);

     if(!err)
     {
      err=dofit(&xstr,0);
      if(xstr.fit.pathsize==0) err=geterror(EPX0);
      else
      {
       vecend(&xstr.fit);

       tr_setshift(&tr,dbox.x0,dbox.y0);
       ctrans=1;
      }
     }
    }

    imdeletescreen(&imhandle);
   }
  }

  if(!err)
  {
   deletecurve(&c,file);
   newcurve(&c,CPATH,&xstr.fit.path,xstr.fit.pathsize,file);

   if(ctrans) transformcurve(&tr,&c,file);
   else       validpath(&c.box,&c.data.path.flags,xstr.fit.path);

   pictp=(pictboxstr*)fob_find(pict,file);
   pictp->pcurve=c; 
   pictp->masked=masked;
  }

  vecclear(&xstr.fit);
 }

 hourglassoff();

 return(err);
}



os_error * xtraceinit(void)
{
 return(fitinit());
}

