/*->c.spell */



#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

#include "os.h"
#include "wimp.h"
#include "sprite.h"
#include "font.h"
#include "bbc.h"
#include "flex.h"


#include "fsx.h"
#include "bf.h"
#include "scanplus.h"
#include "bits.h"
#include "transform.h"
#include "wos.h"
#include "poll.h"
#include "err.h"
#include "deb.h"

#include "etc.h"


#include "str.h"
#include "constants.h"
#include "reslink.h"


#include "obj.h"

#include "file.h"

#include "codes.h"

#include "txchoices.h"

#include "th.h"

#include "dc.h"

#include "spell.h"



/* main spell check file */
/* sort out loading dictionaries etc. for files */


#define DTABLECHUNK (16*sizeof(dlinestr))


dlinestr * dtable;
int        totnames=0;     /* number of dictionaries             */

       int         hpdcno;

       domainstr * ddomain;



static void initline(int line,int user,char * name)
{
 memset(&dtable[line],0,sizeof(dlinestr));
 dtable[line].user=user;
 xzstrncpy(dtable[line].name,name,sizeof(dtable[line].name));
}


static os_error * inittable(void)
{
 totnames=0;
 return(flex_alloc((flex_ptr)&dtable,DTABLECHUNK));
}







/* scan a dictionary directory */

static os_error * loadnamedir(char * dir,int user)
{
 os_error   * err;
 fxstat       f;
 char       * p;
 fsscanhandle	handle;
	int          eof;

 err=fs_startsortscan(dir,TEXT("*.*"),0,&handle);
 if(!err)
 {
  while(1)
  {
   err=fs_nextsortitem(handle,&f,&eof);
			if(err) break;
			if(eof) break;

   p=fs_extension(f.name);
   if(p && !cstrcmp(p,"dic"))
   {
    *(p-1)=0;

    err=flex_chunk((flex_ptr)&dtable,(totnames+1)*sizeof(dlinestr),DTABLECHUNK);
    if(err) break;
    initline(totnames,user,f.name);
    totnames++;
   }
  }
  fs_endsortscan(&handle);
 }
 return(err);
}


/* loads the dictionary names */


static int cmpdc(const void * p1,const void * p2)
{
 int code;

           code=cstrcmp(((dlinestr*)p1)->name,((dlinestr*)p2)->name);
 if(!code) code=((dlinestr*)p2)->user-((dlinestr*)p1)->user;

 return(code);
}




static os_error * loadnames(void)
{
 os_error * err;
 int        user;
 int        r;
 int        w;
 int        same;
 char       temp0[FILENAMELEN];
 char       temp1[FILENAMELEN];

 /* scan the dictionary directory */

 err=NULL;

 totnames=0;

 fs_expandpath(dcpath[0],temp0,sizeof(temp0));
 fs_expandpath(dcpath[1],temp1,sizeof(temp1));
 
 same=!strcmp(temp0,temp1);

 for(user=same;user<2;user++)
 {
  err=loadnamedir(dcpath[user],user);
  if(err) break;
 }

 if(!same)
 {
  qsort(dtable,totnames,sizeof(dlinestr),cmpdc);

  w=0;
  for(r=1;r<totnames;r++)
  {
   if(cstrcmp(dtable[r].name,dtable[w].name))
   {
    w++;
    if(w!=r) dtable[w]=dtable[r];
   }
  }

  totnames=w+1;
 }

 return(err);
}


/* given a dictionary number - load the dictionary */

os_error * loaddcn(int n)
{
 os_error * err;

 err=loaddc(dtable[n].name,&dtable[n].dict,dtable[n].user,ddomain);
 if(!err) dtable[n].loaded=1;
 else     dtable[n].loaded=0;

/* testdict(&dtable[n].dictp); */

 return(err);
}


/* given a dictionary number - remove the dictionary from memory */

static os_error * unloaddcn(int n)
{
 trashdc(&dtable[n].dict,ddomain);
 dtable[n].loaded=0;
 return(NULL);
}


os_error * dccleanup(void)
{
 os_error * err;
 int        i;

 err=NULL;

 for(i=0;i<totnames;i++)
 {
/*dprintf(0,"i %d uses %d loaded %d",i,dtable[i].uses,dtable[i].loaded); */

  if(dtable[i].uses<=0 && dtable[i].loaded) err=unloaddcn(i);
  else
  if(dtable[i].uses>0 && !dtable[i].loaded) err=loaddcn(i);
  if(err) break;
 }

 ob_compactdomain(ddomain);

 return(err);
}


/* given a dictionary name, return its number or -1 */

int finddc(char * name)
{
 int i=0;
 while(i<totnames)
 {
  if(!cstrcmp(dtable[i].name,name)) return(i); 
  else                              i++;
 }
 return(-1);
}


#ifdef NEVER

/* how many dictionaries are there? */

static int nodcs(void)
{
 return(totnames);
}


/* how many dictionaries are there excluding hyphenation dictionary */

static int nodcsexhp(void)
{
 if(hpdcno==-1) return(totnames);
 else           return(totnames-1);
}


#endif

/*****************************************************************************/

os_error * copydcmap(filestr * sfile,filestr * dfile)
{
 os_error      * err;
 spellstr      * spellp;
 int             spell;
 int             i;
 int             dict;

 err=fob_copy((fobjptr)&spellp,dfile,sfile->hdr.spell,sfile);
 if(!err)
 {
  spell=spellp->key;
  dfile->hdr.spell=spell;

  for(i=0;i<spellp->nodc;i++)
  {
   if(spellp->spdict[i].embedded)
   {
    if(!err)
    {
     err=copydc(&dict,dfile->domain,spellp->spdict[i].index,sfile->domain);
    }
    else dict=0;

    spellp=(spellstr*)fob_find(spell,dfile);
    spellp->spdict[i].index=dict;
   }
   else
   {
   /* dtable[spellp->spdict[i].index].uses++; */
   /* - no because uses set by fopen event    */
   /* but not when making style sheets        */
   }
  }
 }

 return(err);
}


os_error * newdcmap(filestr * file)
{
 os_error      * err;
 int             shandle=0; /* init to make compiler happy */
 spellstr      * spell;

 err=fob_create(sizeof(spellstr),file,OSPELL,(fobjptr)&spell);
 if(!err)
 {
  shandle=spell->key;
  file->hdr.spell=shandle;
 }
 return(err);
}


os_error * remfiledc(filestr * file)
{
 os_error * err;
 int        i;
 int        spellk;
 spellstr * spellp;

 err=NULL;

 spellk=file->hdr.spell;
 spellp=(spellstr*)fob_find(spellk,file);

 for(i=0;i<spellp->nodc;i++)
 {
  if(!spellp->spdict[i].embedded)
  {
   dtable[spellp->spdict[i].index].uses--;
  }
 }

 return(err);
}


os_error * addfiledc(filestr * file)
{
 os_error * err;
 int        i;
 int        spellk;
 spellstr * spellp;

 err=NULL;

 spellk=file->hdr.spell;
 spellp=(spellstr*)fob_find(spellk,file);

// cprintf("add nodc %d",spellp->nodc);

 for(i=0;i<spellp->nodc;i++)
 {
  if(!spellp->spdict[i].embedded)
  {
   dtable[spellp->spdict[i].index].uses++;
  }
 }

 return(err);
}



static os_error * checkembedded(filestr * file)
{
 os_error * err;
 int        i;
 int        spellk;
 spellstr * spellp;
 odictstr * dp;
 int        n;
 int        good;

 err=NULL;

 spellk=file->hdr.spell;
 if(spellk)
 {
  spellp=(spellstr*)fob_find(spellk,file);
  if(spellp)
  {
   for(n=i=0;i<spellp->nodc;i++)
   {
    good=0;
    if(spellp->spdict[i].embedded)
    {
     dp=(odictstr*)fob_find(spellp->spdict[i].index,file);
     if(dp)
     {
      if(dp->type==ODICT)
      {
       fob_delete(spellp->spdict[i].index,file);
       cprintf("++ removed embedded dictionary");
       spellp=(spellstr*)fob_find(spellk,file);
      }
      else
      if(dp->type==ODICT2)
      {
       good=1;
      }
     }
    }
    else
    {
     good=1;
    }

    if(good)
    {
     if(i!=n)
     {
      spellp->spdict[n]=spellp->spdict[i];
     }
     n++;
    }
   }

   spellp->nodc=n;
  }
 }

 return(err);
}





static os_error * remapdc(filestr * file)
{
 os_error * err;
 int        i;
 int        j;
 int        n;
 int        spellk;
 spellstr * spellp;

 err=NULL;

 spellk=file->hdr.spell;

 spellp=(spellstr*)fob_find(spellk,file);

 n=0;
 for(i=0;i<spellp->nodc;i++)
 {
  if(spellp->spdict[i].embedded) j=spellp->spdict[i].index;
  else                           j=finddc(spellp->spdict[i].name);

  if(j>=0)
  {
   spellp->spdict[i].index=j;
   if(i!=n)
   {
    spellp->spdict[n]=spellp->spdict[i];
   }
   n++;
  }
 }

 spellp->nodc=n;

 return(err);
}



static os_error * remapalldc(void)
{
 filestr * file;

 file=firstfile;

 while(file)
 {
  remapdc(file);
  file=file->next;
 }

 return(NULL);
}




static os_error * scrapignore(filestr * file)
{
 spellstr * spellp;
 int        i;
 int        dict;

/* dprintf(2,"scrap ignore %d",saveignore); */

 if(!saveignore)
 {
  spellp=(spellstr*)fob_find(file->hdr.spell,file);

  for(i=0;i<spellp->nodc;i++)
  {
   if(spellp->spdict[i].ignore)
   {
    dict=spellp->spdict[i].index;
    spellp->spdict[i]=spellp->spdict[spellp->nodc-1];
    spellp->nodc--;

    if(spellp->spdict[i].embedded)
    {
     trashdc(&dict,file->domain);
    }

    break;
   }
  }
 }

 return(NULL);
}



/* called when file is opened/closed */

static os_error * spevent(int eventn,int user,filestr * file,int data1)
{
// cprintf("spevent %d",eventn);
 
 if(eventn==EVENT_FCLOSE)
 {
  remfiledc(file);
  dccleanup();
 }
 else
 if(eventn==EVENT_FOPEN)
 {
  checkembedded(file);
  remapdc(file);
  addfiledc(file);
  dccleanup();
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(data1);
}



static os_error * spsave(int eventn,int user,filestr * file,int data1)
{
 return(scrapignore(file));

 return(NULL);
 USE(eventn);
 USE(user);
 USE(data1);
}


/*****************************************************************************/

os_error * deletedict(int n)
{
 os_error * err;
 char       string[FILENAMELEN];
 int        i;

 strcpy(string,dcpath[dtable[n].user]);
 strcat(string,"\\");
 strcat(string,dtable[n].name);
 strcat(string,".dic");

 err=fs_delete(string);
 if(!err)
 {
  unloaddcn(n);

  if(n==hpdcno) hpdcno=-1;
  else
  if(hpdcno>n)  hpdcno--;

  for(i=n;i<(totnames-1);i++) 
  {
   dtable[i]=dtable[i+1];
  }

  totnames--;
 }

 remapalldc();

 return(err);
}


/* called from menu to create a dictionary */

os_error * createdict(char * string,int * index)
{
 os_error * err;
 int        i;
 int        n;

 err=NULL;

 for(i=0;i<totnames;i++)
 {
  if(!cstrcmp(dtable[i].name,string))
  {
   err=geterror(ESP01);
   break;
  }
 }

 if(!err)
 {
  for(n=0;n<totnames;n++)
  {
   if(strcmp(dtable[n].name,string)>0) break;
  }

  err=flex_chunk((flex_ptr)&dtable,(totnames+1)*sizeof(dlinestr),DTABLECHUNK);

  if(!err)
  {
   totnames++;
   for(i=totnames-1;i>n;i--) 
   {
    dtable[i]=dtable[i-1];
   }

   initline(n,1,string);
   dtable[n].loaded=1;

   if(hpdcno>=n) hpdcno++;

   err=createdc(&dtable[n].dict,ddomain);
   if(!err)
   {
    err=savedc(dtable[n].name,dtable[n].dict,dtable[n].user,ddomain);
    if(err) deletedict(n);
    else    
    {
     remapalldc();
     *index=n;
    }
   }
  }
 }

 return(err);
}

/****************************************************************************/


os_error * spsavechanges(void)
{
 os_error * err;
 int        i;

 err=NULL;

 for(i=0;i<totnames;i++)
 {
  if(dtable[i].modded)
  {
   err=savedc(dtable[i].name,dtable[i].dict,dtable[i].user,ddomain);
   if(err) break;
   dtable[i].modded=0;
  }
 }

 return(err);
}


os_error * spscrapchanges(void)
{
 os_error * err;
 int        i;

 err=NULL;

 for(i=0;i<totnames;i++) 
 {
  if(dtable[i].modded)
  {
   trashdc(&dtable[i].dict,ddomain);
   err=loaddcn(i);
   if(err) break;
   dtable[i].modded=0;
  }
 }

 return(err);
}


int spchanges(void)
{
 int i;
 for(i=0;i<totnames;i++) if(dtable[i].modded) return(1);
 return(0);
}


/*****************************************************************************/


void spstripstring(char * dest,char * src,int destsize)
{
 int    c;
 char * limit;

 limit=dest+destsize;

 do
 {
  c=*src++;
  if(c==NBHYPH) c='-';
  else
  if(c==145) c='\'';
  *dest++=(char)c;
  if(dest>=limit) break;
 } while(c);

 *(limit-1)=0;

}





/*****************************************************************************/
                     /* tie into hyphenation code */

/* load hyphenation dictionary and lock into memory */

static os_error * inithpdc(void)
{
 os_error * err;

 err=NULL;

 hpdcno=finddc("!HyphDict");
 if(hpdcno!=-1)
 {
  err=loaddcn(hpdcno);
  if(!err) dtable[hpdcno].uses=255;
 }

 return(err);
}


/* minlen == min number of chars before hyphen */

int checkhpdc(char * string,char * result,int minlen)
{
 word words;
 int  flags=0x30000+(minlen<<24);

 if(hpdcno>=0 && findwordabs(string,dtable[hpdcno].dict,&words,&flags,ddomain))
 {
  strcpy(result,words.string);
  return(1);
 }

 return(0);
}

/*****************************************************************************/


os_error * spellinit(void)
{
 os_error   * err;

 cprintf("spell init");

          err=inittable();
	if(!err) err=dcpathinit();
 if(!err) err=loadnames();
 if(!err) 
 {
  obufferstr * buffer; /* create a dummy object so domain always has one in */

  err=ob_createdomain(&ddomain);
  ob_create(sizeof(obufferstr),ddomain,OBUFFER,(objptr)&buffer);
 }

 if(!err) err=addevent(EVENT_FCLOSE,(eventfn)spevent,0);
 if(!err) err=addevent(EVENT_FOPEN,(eventfn)spevent,0);
 if(!err) err=addevent(EVENT_FPRESAVE,(eventfn)spsave,0);

 if(!err) err=inithpdc();

 if(!err) thinit();

 return(err);
}

