/*->c.tx */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  
#include <ctype.h>
#include <assert.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"


#include "transform.h"

#include "etc.h"
#include "err.h"
#include "wos.h"
#include "deb.h"

#include "str.h"
#include "constants.h"

#include "file.h"
#include "undo.h"

#include "object.h"


#include "codes.h"

#include "txstr.h"
#include "vl.h"
#include "style.h"
#include "fl.h"
#include "fl2.h"

#include "act.h"
#include "bms.h"


#include "tx.h"



/*                                                                                 handles buffers of text at a low level 
*/


#define TXCHUNK 0x400


os_error * txsourcenew(int * key,filestr * file)
{
 os_error  * err;
 sourcestr * sourcep;

 err=fob_create(sizeof(sourcestr),file,OSOURCE,(fobjptr)&sourcep);
 if(!err)
 {
  sourcep->len=0;
  sourcep->maxlen=0;
  sourcep->goffset=0;
  *key=sourcep->key;
 }
 return(err);
}



os_error * txsourcecopy(int src,int * key,filestr * sfile,filestr * dfile)
{
 os_error  * err;
 sourcestr * sourcep;


 err=fob_copy((fobjptr)&sourcep,dfile,src,sfile);

 *key=sourcep->key;


 return(err);
}





static void setgapto(sourcestr * source,int offset)
{
 char * buff;
 int    len;
 int    maxlen;
 int    goffset;
 int    gapsize;

 buff=source->data;
 len=source->len;
 maxlen=source->maxlen;
 goffset=source->goffset;
 gapsize=maxlen-len;

 if(offset>goffset) offset+=gapsize;

 if(goffset>offset)
 {
  memmove(buff+offset+gapsize,buff+offset,goffset-offset);
  source->goffset=offset;
 }
 else
 if(offset>goffset)
 {
  memmove(buff+goffset,buff+goffset+gapsize,offset-goffset-gapsize);
  source->goffset+=offset-goffset-gapsize;
 }
}




os_error * txsourceclose(textsrcstr * textsrc,filestr * file)
{
 os_error  * err;
 int         free;
 sourcestr * sourcep;
 int         len;

 err=NULL;

 len=textsrc->sourcep->len;

 free=textsrc->sourcep->maxlen-len;

 if(free)
 {
  setgapto(textsrc->sourcep,len);
  err=fob_change(textsrc->source,file,
                   sizeof(sourcestr)+len,(fobjptr)&sourcep);
  textsrc->sourcep->maxlen=len;
 }

 return(err);
}





os_error * txsourcedel(int key,filestr * file,int contents)
{
 os_error * err;
 sourcestr * sourcep;
 int        c;
 char     * p;
 int        len;
 int        maxlen;
 int        bits;
 int        n;
 int        offset;

 err=NULL;

 if(key)
 {
  bmsourcedel(key,file);

  if(contents)
  {
   err=fob_lock((fobjptr)&sourcep,key,file);
   if(!err)
   {
    maxlen=sourcep->len;
    offset=0;

    while(offset<maxlen)
    {
     p=txptr(sourcep,offset);

     c=*p;
     bits=codex[c];
     len=(bits & CODELEN);

     if(c==SNO)
     {
      n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
      remabsstyle(file,n);
     }
     else
     if(c==EOB)
     {
      n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);
      objectdelete((objhdrstr*)fob_find(n,file),DELETESCRAP,file);
     }

     offset+=len;
    }

    err=fob_unlock((fobjptr)&sourcep,file);
   }
  }

  return(fob_delete(key,file));
 }
 else
  return(NULL);
}




static os_error * txsourceinserts(int n,textsrcstr * textsrc,filestr * file)
{
 os_error  * err;
 int         free;
 int         xnew;
 sourcestr * sourcep;
 int         topgapoffset;

 err=NULL;

 if(n)
 {
  flsourcechange(textsrc,file,n);
  bmsourcechange(textsrc,file,n);

  setgapto(textsrc->sourcep,textsrc->offset);
  free=textsrc->sourcep->maxlen-textsrc->sourcep->len;

  if(free<n)
  {
   xnew=n-free+TXCHUNK;
   err=fob_change(textsrc->source,file,
                     sizeof(sourcestr)+textsrc->sourcep->maxlen+xnew,
                                                       (fobjptr)&sourcep);
   if(!err)
   {
    topgapoffset=textsrc->offset+free;

    memmove(sourcep->data+topgapoffset+xnew,
            sourcep->data+topgapoffset,
            sourcep->maxlen-topgapoffset);

    sourcep->maxlen+=xnew;
    free=sourcep->maxlen-sourcep->len;
   }
  }

  if(!err)
  {
   textsrc->sourcep->len+=n;
   free-=n;
   textsrc->sourcep->goffset+=n;
   if(free>0) *(textsrc->sourcep->data+textsrc->sourcep->goffset)=GCODE;
  }
 }

 return(err);
}



os_error * txsourceinsert(char * data,int n,
                                      textsrcstr * textsrc,filestr * file)
{
 os_error * err;

 err=txsourceinserts(n,textsrc,file);
 if(!err) memmove(textsrc->sourcep->data+textsrc->offset,data,n);

 return(err);
}



os_error * txsourceinsertsn(int sn,int dup,textsrcstr * textsrc,filestr * file)
{
 char       sbuff[SNLEN];

 if(dup) dupabsstyle(file,sn);

 sbuff[0]=SNO;
 sbuff[1]=(char)sn;
 sbuff[2]=(char)(sn>>8);
 sbuff[3]=(char)(sn>>16);

 return(txsourceinsert(sbuff,SNLEN,textsrc,file));
}


static os_error * txsourceinsertb(int txobject,int n,
                                        textsrcstr * textsrc,filestr * file)
{
 os_error   * err;
 txscrapstr * txs;

 err=txsourceinserts(n,textsrc,file);
 if(!err)
 {
  txs=(txscrapstr*)fob_find(txobject,file);
  memmove(textsrc->sourcep->data+textsrc->offset,txs->stuff,n);
 }

 return(err);
}



os_error * txundosourceinsert(char * data,int n,
                                          textsrcstr * textsrc,filestr * file)
{
 txmodstr txmod;

 txmod.offset=textsrc->offset;
 txmod.size=n;
 undowritefn(file,UDELETE,&txmod,sizeof(txmodstr));

 return(txsourceinsert(data,n,textsrc,file));
}



os_error * txundosourceinsertsn(int sn,textsrcstr * textsrc,filestr * file)
{
 char sbuff[SNLEN];

 dupabsstyle(file,sn);

 sbuff[0]=SNO;
 sbuff[1]=(char)sn;
 sbuff[2]=(char)(sn>>8);
 sbuff[3]=(char)(sn>>16);

 return(txundosourceinsert(sbuff,SNLEN,textsrc,file));
}




os_error * txsourceremove(int n,textsrcstr * textsrc,filestr * file)
{
 if(n)
 {
  flsourcechange(textsrc,file,-n);
  bmsourcechange(textsrc,file,-n);

  setgapto(textsrc->sourcep,textsrc->offset);
  textsrc->sourcep->len-=n;
  *(textsrc->sourcep->data+textsrc->offset)=GCODE;
 }

 return(NULL);
 USE(file);
}



#define UBSIZE 512


typedef struct txmodbstr
{
 txmodstr   txmod;
 char       buff[UBSIZE];
} txmodbstr;



static os_error * txundosourceremove(int n,textsrcstr * textsrc,filestr * file)
{
 os_error * err;
 txmodbstr  txmodb;
 int        offset;
 int        size;
 int        maxsize;
 int        len;
 char     * p;
 char     * q;


 txmodb.txmod.offset=textsrc->offset;
 size=n;

 setgapto(textsrc->sourcep,textsrc->offset);

 offset=textsrc->offset;
 offset+=textsrc->sourcep->maxlen-textsrc->sourcep->len;

 while(size>0)
 {
  p=q=textsrc->sourcep->data+offset;
  txmodb.txmod.size=0;
  maxsize=MIN(UBSIZE,size);
  len=0;

  while(txmodb.txmod.size<maxsize)
  {
   len=(codex[*p] & CODELEN);
   txmodb.txmod.size+=len;
   p+=len;
  }
  if(txmodb.txmod.size>maxsize) txmodb.txmod.size-=len;

  XMEMCPYX(txmodb.buff,q,txmodb.txmod.size);  

  undowritefn2(file,UINSERT,&txmodb,sizeof(txmodstr),txmodb.txmod.size);

  /* txmodb.txmod.offset+=txmodb.txmod.size;         */
  /* because on undo reads will be in opposite order */
  /* means text will be shifted for each block read  */

  offset+=txmodb.txmod.size;
  size-=txmodb.txmod.size;
 }

 err=txsourceremove(n,textsrc,file);

 return(err);
}



static os_error * txundosourceremoveb(int n,textsrcstr * textsrc,filestr *file)
{
 os_error   * err;
 int          key;
 txscrapstr * txs;
 int          offset;

 setgapto(textsrc->sourcep,textsrc->offset);

 err=fob_create(sizeof(txscrapstr)+n,file,OTXSCRAP,(fobjptr)&txs);
 if(!err)
 {
  key=txs->key;
  offset=textsrc->offset;
  txs->txmod.size=n;
  txs->txmod.offset=offset;

  offset+=textsrc->sourcep->maxlen-textsrc->sourcep->len;

  XMEMCPYX(txs->stuff,(textsrc->sourcep)->data+offset,n);

  undowritefn(file,UINSERTB,&key,sizeof(int));

  err=txsourceremove(n,textsrc,file);
 }

 return(err);
}




os_error * txundosourceremovex(int n,textsrcstr * textsrc,filestr * file)
{
 if(n>=BSSPLIT) return(txundosourceremoveb(n,textsrc,file));
 else           return(txundosourceremove(n,textsrc,file));
}



/*
os_error * txundosourcereplace(int delete,int insert,char * data,
                                           textsrcstr * textsrc,filestr * file)
{
 os_error * err;

          err=txundosourceremove(delete,textsrc,file);
 if(!err) err=txundosourceinsert(data,insert,textsrc,file);

 return(err);
}
*/


os_error * txsourcereplace(int xdelete,int insert,char * data,
                                           textsrcstr * textsrc,filestr * file)
{
 os_error * err;

          err=txsourceremove(xdelete,textsrc,file);
 if(!err) err=txsourceinsert(data,insert,textsrc,file);

 return(err);
}




static textsrcstr   undotextsrc;
static int          undofoffset;
static int          undoxoffset;
static int          undodelta;


void txundostart(textsrcstr * txsrc,filestr * file)
{
/* dprintf(0,"txundostart"); */

 undotextsrc=*txsrc;
 txlock(&undotextsrc,file);
 undofoffset=-1;
 undoxoffset=-1;
 undodelta=0;
}


void txundoend(int * loffset,int * hoffset,int * delta,filestr * file)
{
/* dprintf(0,"txundoend"); */

 txunlock(&undotextsrc,file);

 if(loffset) *loffset=undofoffset;
 if(hoffset) *hoffset=undoxoffset;
 if(delta)   *delta=undodelta;
}




/* tagged to UINSERT */

os_error * uninsert(filestr * file) 
{
 os_error * err;
 txmodbstr  txmodb;

 err=undoread(file,&txmodb.txmod,sizeof(txmodstr));

 if(undofoffset<0) undofoffset=txmodb.txmod.offset;
 else              undofoffset=MIN(undofoffset,txmodb.txmod.offset);

 if(undoxoffset<0) undoxoffset=txmodb.txmod.offset+txmodb.txmod.size;
 else              undoxoffset=
                     MAX(undoxoffset,(txmodb.txmod.offset+txmodb.txmod.size));

 undodelta+=txmodb.txmod.size;

 undoread(file,txmodb.buff,txmodb.txmod.size);
 undotextsrc.offset=txmodb.txmod.offset;

 txsourceinsert(txmodb.buff,txmodb.txmod.size,&undotextsrc,file);

 undowritefn(file,UDELETE,&txmodb.txmod,sizeof(txmodstr));

 return(err);
}




os_error * uninsertscrap(filestr * file)
{
 txmodbstr  txmodb;
 char     * p;
 int        n;
 int        c;
 int        bits;
 int        len;

 undoread(file,&txmodb.txmod,sizeof(txmodstr));
 undoread(file,txmodb.buff,txmodb.txmod.size);
 p=txmodb.buff;

/* dprintf(0,"size=%d",txmodb.txmod.size); */

 while(txmodb.txmod.size>0)
 {
  c=*p;
  bits=codex[c];
  len=(bits & CODELEN);

/* dprintf(0,"c=%d",c); */

  if(c==SNO)
  {
   n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
   remabsstyle(file,n);
  }
  else
  if(c==EOB)
  {
   n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);
   objectdelete((objhdrstr*)fob_find(n,file),DELETESCRAP,file);
  }

  p+=len;
  txmodb.txmod.size-=len;
 }

 return(NULL);
}




/* tagged to UINSERTB */

os_error * uninsertb(filestr * file) 
{
 os_error   * err;
 int          txobject;
 txmodstr     txmod;
 txscrapstr * txs;

 err=undoread(file,&txobject,sizeof(int));

 txs=(txscrapstr*)fob_find(txobject,file);
 txmod=txs->txmod;

 if(undofoffset<0) undofoffset=txmod.offset;
 else              undofoffset=MIN(undofoffset,txmod.offset);

 if(undoxoffset<0) undoxoffset=txmod.offset+txmod.size;
 else              undoxoffset=
                     MAX(undoxoffset,(txmod.offset+txmod.size));

 undodelta+=txmod.size;

 undotextsrc.offset=txmod.offset;

 txsourceinsertb(txobject,txmod.size,&undotextsrc,file);

 undowritefn(file,UDELETEB,&txmod,sizeof(txmodstr));

 fob_delete(txobject,file);

 return(err);
}



os_error * uninsertscrapb(filestr * file)
{
 os_error   * err;
 int          txobject;
 int          size;
 txscrapstr * txs;
 int          c;
 int          len;
 int          bits;
 int          n;
 char       * p;


/* dprintf(0,"uninsertscrapb!"); */

 err=undoread(file,&txobject,sizeof(int));

 txs=(txscrapstr*)fob_find(txobject,file);
 size=txs->txmod.size;
 p=txs->stuff;

 while(size>0)
 {
  c=*p;
  bits=codex[c];
  len=(bits & CODELEN);

  if(c==SNO || c==EOB)
  {
   if(c==SNO)
   {
    n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
    remabsstyle(file,n);
   }
   else
   if(c==EOB)
   {
    n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);
    objectdelete((objhdrstr*)fob_find(n,file),DELETESCRAP,file);
   }

   txs=(txscrapstr*)fob_find(txobject,file);
   p=txs->stuff+txs->txmod.size-size;
  }

  p+=len;
  size-=len;
 }

 fob_delete(txobject,file);

 return(err);
}






/* tagged to UDELETE */


static os_error * undeletesub(filestr * file,int b)
{
 txmodstr   txmod;

 undoread(file,&txmod,sizeof(txmodstr));

 if(undofoffset<0) undofoffset=txmod.offset;
 else              undofoffset=MIN(undofoffset,txmod.offset);

 if(undoxoffset<0) undoxoffset=txmod.offset;
 else              undoxoffset=MAX(undoxoffset,txmod.offset);

 undodelta-=txmod.size;

 undotextsrc.offset=txmod.offset;

 if(b)  return(txundosourceremoveb(txmod.size,&undotextsrc,file));
 else   return(txundosourceremove(txmod.size,&undotextsrc,file));
}



os_error * undelete(filestr * file)
{
 return(undeletesub(file,0));
}


/* tagged to UDELETEB */

os_error * undeleteb(filestr * file)
{
 return(undeletesub(file,1));
}


/****************************************************************************/

#ifdef NEVER


/* old versions */


os_error * txundosourceremove(int n,textsrcstr * textsrc,filestr * file)
{
 os_error * err;
 txmodstr   txmod;
 char       buff[UBSIZE];
 int        bn;
 int        i;
 int        last;
 int        offset;

 txmod.offset=textsrc->offset;
 txmod.size=n;
 undowritefn(file,UINSERT,&txmod,sizeof(txmodstr));

 bn=n/UBSIZE;
 last=n-bn*UBSIZE;

 setgapto(textsrc->sourcep,textsrc->offset);

 offset=textsrc->offset;
 offset+=textsrc->sourcep->maxlen-textsrc->sourcep->len;

 for(i=0;i<bn;i++)
 {
  XMEMCPYX(buff,textsrc->sourcep->data+offset,UBSIZE);
  undowrite(file,buff,UBSIZE);
  offset+=UBSIZE;
 }
 if(last)
 {
  XMEMCPYX(buff,textsrc->sourcep->data+offset,last);
  undowrite(file,buff,last);
 }

 err=txsourceremove(n,textsrc,file);

 return(err);
}




os_error * uninsertscrap(filestr * file)
{
 char       buff[UBSIZE]; 
 txmodstr   txmod;
 int        read;
 int        offset;
 char     * p;
 int        n;
 int        c;
 int        bits;
 int        len;


 undoread(file,&txmod,sizeof(txmodstr));
 offset=UBSIZE;

 while(txmod.size>0)
 {
  read=MIN(txmod.size,UBSIZE-(UBSIZE-offset));
  undoread(file,buff+(UBSIZE-offset),read);
  txmod.size-=read;
  read+=(UBSIZE-offset);

  p=buff;
  offset=0;
  while(offset<read)
  {
   c=*p;
   bits=codex[c];
   len=(bits & CODELEN);

   if((len+offset)>read)
   {
    XMEMCPYX(buff,p,read-offset);
    break;
   }
   else
   {
    if(c==SNO)
    {
     n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
     remabsstyle(file,n);
    }
    else
    if(c==EOB)
    {
     n=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);
     objectdelete((objhdrstr*)fob_find(n,file),DELETESCRAP,file);
    }

    offset+=len;
    p+=len;
   }
  }
 }

 return(NULL);
}




os_error * uninsert(filestr * file) 
{
 os_error * err;
 txmodstr   txmod;
 int        bn;
 int        last;
 char       buff[UBSIZE];
 int        offset;
 int        i;


 err=undoread(file,&txmod,sizeof(txmodstr));

 if(undofoffset<0) undofoffset=txmod.offset;
 else              undofoffset=MIN(undofoffset,txmod.offset);

 if(undoxoffset<0) undoxoffset=txmod.offset+txmod.size;
 else              undoxoffset=MAX(undoxoffset,(txmod.offset+txmod.size));

 undodelta+=txmod.size;

 undowritefn(file,UDELETE,&txmod,sizeof(txmodstr));

 bn=txmod.size/UBSIZE;
 last=txmod.size-bn*UBSIZE;

 offset=txmod.offset;

 for(i=0;i<bn;i++)
 {
  undoread(file,buff,UBSIZE);
  undotextsrc.offset=offset;
  txsourceinsert(buff,UBSIZE,&undotextsrc,file);
  offset+=UBSIZE;
 }
 if(last)
 {
  undoread(file,buff,last);
  undotextsrc.offset=offset;
  txsourceinsert(buff,last,&undotextsrc,file);
 }

 return(err);
}





#endif


/****************************************************************************/





/* tagged to USSBLOCK */

os_error * ussblockscrap(filestr * file)
{
 undostylestr uss;
 int          i;
 int          style;

 undoread(file,&uss.ns,sizeof(int));
 undoread(file,&uss.usmap,uss.ns*sizeof(usmapstr));

 for(i=0;i<uss.ns;i++)
 {
  style=uss.usmap[i].style;
  remabsstyle(file,style);
 }

 return(NULL);
}



os_error * ussblock(filestr * file)
{
 undostylestr usmap;
 undostylestr usmap2;
 int      i;
 char     temp[4];
 int      style;
 int      offset;
 int      n;
 char   * p;

 undoread(file,&usmap.ns,sizeof(int));
 undoread(file,&usmap.usmap,usmap.ns*sizeof(usmapstr));

/* dprintf(0,"read ns=%d",usmap.ns); */

 usmap2.ns=usmap.ns;

 for(n=0,i=usmap.ns-1;i>=0;i--,n++)
 {
  style=usmap.usmap[i].style;
  offset=usmap.usmap[i].offset;

  if(undofoffset<0) undofoffset=offset;
  else              undofoffset=MIN(undofoffset,offset);

  if(undoxoffset<0) undoxoffset=offset+SNLEN;
  else              undoxoffset=MAX(undoxoffset,(offset+SNLEN));


  usmap2.usmap[n].offset=offset;

  p=txptr(undotextsrc.sourcep,usmap.usmap[i].offset);
  if(*p==SNO)
  {
   usmap2.usmap[n].style=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);

   if(usmap.usmap[i].style>0)
   {
    *(p+1)=(char)style;
    *(p+2)=(char)(style>>8);
    *(p+3)=(char)(style>>16);
   }
   else
   {
    undotextsrc.offset=offset;
    txsourceremove(SNLEN,&undotextsrc,file);
    undodelta-=SNLEN;
   }
  }
  else
  {
   usmap2.usmap[n].style=0;

   if(usmap.usmap[i].style>0)
   {
    temp[0]=SNO;
    temp[1]=(char)style;
    temp[2]=(char)(style>>8);
    temp[3]=(char)(style>>16);

    undotextsrc.offset=offset;
    txsourceinsert(temp,SNLEN,&undotextsrc,file);
    undodelta+=SNLEN;
   }
  }
 }

 undowritefn2(file,USSBLOCK,&usmap2,sizeof(int),usmap2.ns*sizeof(usmapstr));

/* dprintf(0,"write ns=%d",usmap2.ns); */

/*
 undowritefn(file,USSBLOCK,&usmap2.ns,sizeof(int));
 undowrite(file,&usmap2.usmap,usmap2.ns*sizeof(usmapstr));
*/


 return(NULL);
}




os_error * undostylemapstart(filestr * file,undostylestr * uss)
{
 uss->ns=0;
 return(NULL);
 USE(file);
}


os_error * undostylemapend(filestr * file,undostylestr * uss)
{
 if(uss->ns)
 {
  undowritefn(file,USSBLOCK,&uss->ns,sizeof(int));
  undowrite(file,&uss->usmap,uss->ns*sizeof(usmapstr));
  uss->ns=0;
 }

 return(NULL);
}


os_error * undostylemapadd(filestr * file,undostylestr * uss,int offset,int sn)
{
 uss->usmap[uss->ns].offset=offset;
 uss->usmap[uss->ns++].style=sn;
 if(uss->ns>=USSMAX) undostylemapend(file,uss);

 return(NULL);
}




/************************************************************************/


os_error * txcopyblock(textsrcstr * textsrc,textsrcstr * textsrc2,
                       filestr * file,filestr * file2,int n)
{
 os_error  * err;
 int         free;
 int         xnew;
 sourcestr * sourcep;
 int         g2;
 int         g2size;
 int         n1;

 err=NULL;
 setgapto(textsrc->sourcep,textsrc->offset);
 free=textsrc->sourcep->maxlen-textsrc->sourcep->len;

 if(free<n)
 {
  xnew=n-free+TXCHUNK;
  err=fob_change(textsrc->source,file,
                        sizeof(sourcestr)+textsrc->sourcep->maxlen+xnew,
                                                      (fobjptr)&sourcep);
  if(!err)
  {
   memmove(sourcep->data+textsrc->offset+free+xnew,
           sourcep->data+textsrc->offset+free,
           sourcep->maxlen-(textsrc->offset+free));
   sourcep->maxlen+=xnew;
   free=sourcep->maxlen-sourcep->len;
  }
 }

 if(!err)
 {
  flsourcechange(textsrc,file,n);
  bmsourcechange(textsrc,file,n);

  g2=(textsrc2->sourcep)->goffset;
  if(textsrc2->offset<g2 && (textsrc2->offset+n)>g2)
  {
   g2size=(textsrc2->sourcep)->maxlen-(textsrc2->sourcep)->len;
   n1=g2-textsrc2->offset;

   memmove(textsrc->sourcep->data+textsrc->offset,
                                  textsrc2->sourcep->data+textsrc2->offset,n1);

   memmove(textsrc->sourcep->data+(textsrc->offset+n1),
                                  textsrc2->sourcep->data+g2+g2size,n-n1);

  }
  else
  {
   if(textsrc2->offset>=g2)
   {
    memmove(textsrc->sourcep->data+textsrc->offset,
            textsrc2->sourcep->data+textsrc2->offset+
            (textsrc2->sourcep)->maxlen-(textsrc2->sourcep)->len,n); 
   }
   else
   {
    memmove(textsrc->sourcep->data+textsrc->offset,
            textsrc2->sourcep->data+textsrc2->offset,n); 
   }
  }

  textsrc->sourcep->len+=n;
  free-=n;
  textsrc->sourcep->goffset+=n;
  if(free>0) *(textsrc->sourcep->data+textsrc->sourcep->goffset)=GCODE;
 }

 return(err);

 USE(file2);
}


/* assume all in one file for now */

os_error * txmoveblock(textsrcstr * d,textsrcstr * s,filestr * file,int n)
{
 os_error  * err;
 int         offset;

 err=txcopyblock(d,s,file,file,n);
 if(!err)
 {
  offset=s->offset;
  if(s->source==d->source && s->offset>=d->offset) offset+=n;

  flsourcechange(d,file,-n);
  bmsourcechange(d,file,-n);

  setgapto(s->sourcep,offset);
  s->sourcep->len-=n;
  *(s->sourcep->data+offset)=GCODE;
 }

 return(err);
}




os_error * txlock(textsrcstr * textsrc,filestr * file)
{
 os_error * err;
 texsrc   * texsp;

/* dprintf(0,"lock in"); */

 err=fob_lock((fobjptr)&textsrc->textp,textsrc->text,file);
 if(!err)
 {
  texsp=(texsrc*)(((char*)textsrc->textp)+sizeof(textboxstr));
  texsp+=textsrc->u;

  textsrc->source=texsp->source;
  textsrc->vsource=texsp->vsource;

  err=fob_lock((fobjptr)&textsrc->sourcep,textsrc->source,file);
  if(!err)
  {
   err=fob_lock((fobjptr)&textsrc->vsourcep,textsrc->vsource,file);
   if(err)
   {
    fob_unlock((fobjptr)&textsrc->sourcep,file);
      cprintf("fail on vsource %d",textsrc->vsource);
   }
  }
  else
  {
   cprintf("fail on source %d",textsrc->source);
   /* dprintf(0,"u=%d repeat=%d",textsrc->u,textsrc->textp->data.repeat); */
   fob_unlock((fobjptr)&textsrc->textp,file);
  }
 }

/* dprintf(0,"lock ex"); */

 return(err);
}


os_error * txunlock(textsrcstr * textsrc,filestr * file)
{
 os_error * err;

          err=fob_unlock((fobjptr)&textsrc->textp,file);
 if(!err) err=fob_unlock((fobjptr)&textsrc->sourcep,file);
 if(!err) err=fob_unlock((fobjptr)&textsrc->vsourcep,file);
 return(err);
}



os_error * txclosesrc(textsrcstr * textsrc,filestr * file)
{
 os_error * err;

 err=txlock(textsrc,file);
 if(!err)
 {
  txsourceclose(textsrc,file);
  vsourceclose(textsrc,file);
  txunlock(textsrc,file);
 }

 return(err);
}



int txcodelen(int code,sourcestr * source)
{
 if(code==GCODE) return(source->maxlen-source->len);
 else            return(codex[code] & CODELEN);
}


char * txptr(sourcestr * source,int offset)
{

#ifdef _DEBUG

 assert(offset<source->len && offset>=0);
 assert(source->goffset<=source->len && source->goffset>=0);

#else


#endif

 if(offset>=source->goffset) offset+=source->maxlen-source->len;
 return(source->data+offset);
}



char * txptrlt(sourcestr * source,int offset)
{

 if(offset>source->len) offset=source->len;

#ifdef _DEBUG

 assert(offset>=0);
 assert(source->goffset<=source->len && source->goffset>=0);

#else


#endif


 if(offset>=source->goffset) offset+=source->maxlen-source->len;
 return(source->data+offset);
}




os_error * txfindeobs(int object,int source,filestr * file,int * offset)
{
 os_error  * err;
 sourcestr * sourcep;
 int         off;
 char      * p;
 int         o;
 int         c;
 int         bits;


 err=NULL;
 *offset=-1;

 err=fob_lock((fobjptr)&sourcep,source,file);
 if(!err)
 {
  off=0;

  while(1)
  {
   p=txptr(sourcep,off);
   c=*p;
   bits=codex[c];
   if(c==EOB)
   {
    o=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);

    if(o==object)
    {
     *offset=off;
     break;
    }
   }
   else
   if(c==EOT) break;

   off+=bits & CODELEN;
  }
  fob_unlock((fobjptr)&sourcep,file);
 }

 return(err);
}



os_error * txfindeob(int object,int text,filestr * file,int * offset,int * up)
{
 os_error   * err;
 textboxstr * textp;
 texsrc     * texsp;
 int          header;
 int          footer;
 int          u;

 err=NULL;

 textp=(textboxstr*)fob_find(text,file);

 header=textp->data.header;
 footer=textp->data.footer;

 for(u=0;u<3;u++)
 {
  if(u==0 || (u==1 && header) || (u==2 && footer))
  {
   texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
   err=txfindeobs(object,texsp[u].source,file,offset);
   if(*offset>=0) break;
  }
 }

 *up=u;

 return(err);
}






int sourcesize(int src,filestr * file)
{
 sourcestr * sourcep;

 sourcep=(sourcestr*)fob_find(src,file);

 return(sourcep->len);
}


