/*->c.Drawfiles */

/* -> c.DrawFiles
 *
 * DrawFile module, internal code
 * History:
 * Version 0.1: 15 Feb 89, DAHE: created
 *
 * This file consists of code taken from the !Draw version of files c.draw
 * and c.drawDispl for use in the DrawFile module (release 0.2).
 *
 */


#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "os.h"
#include "bbc.h"
#include "sprite.h"
#include "font.h"
#include "wimp.h"

#include "transform.h"
#include "xprocess.h"
#include "xmath.h"
#include "colour.h"
#include "sprx.h"
#include "drawmod.h"
#include "imlib.h"
#include "xxjpeg.h"
#include "deb.h"

#include "DrawFile1.h"
#include "DrawLevel0.h"
#include "DrawFile2.h"


/*=========================================================================*/
/* Code from c.draw */

/* The following are needed to shut the preprocessor up */

/* Return the address of the first path element in a path object */
/* this starts either after the path header (no dash pattern)    */
/* or after the dash pattern                                     */
 
path_movestr *address_pathstart(draw_objptr hdrptr)
{
  if (hdrptr.pathp->pathstyle.joincapwind & packmask_dashed)
    return (path_movestr*)
           &(hdrptr.pathp->data.dashelements[hdrptr.pathp->data.dashcount]);
  else
    return (path_movestr*)&hdrptr.pathp->data;
}


/* Return the address of the dash pattern of a path object */
/* or zero                                                 */

draw_dashstr *address_dashstart(draw_objptr hdrptr)
{
  if (hdrptr.pathp->pathstyle.joincapwind & packmask_dashed)
    return &hdrptr.pathp->data;
  else
    return 0;
}


/*=========================================================================*/
/* Code from c.drawDispl */

#define visible(A)  ((A.objhdrp->bbox.x0<=clip_x1) &&   \
                     (A.objhdrp->bbox.y0<=clip_y1) &&   \
                     (A.objhdrp->bbox.x1>=clip_x0) &&   \
                     (A.objhdrp->bbox.y1>=clip_y0)      \
                    )



/* ScaleT - for text position coords ('font_paint' takes graphics coords)  */
/* ScaleS - for sprite position coords ('put_sprite takes graphics coords) */
/* ScaleF - T.B.A                                                          */

#define xscaleFR(A)       ((A)/40)
#define yscaleFR(A)       ((A)/40)


/* GCol(action,colour) where colour is BBGGRRxx (foreground colour) */



static os_error *displ_settruecol(int action,draw_coltyp colour)
{
 return(Draw_drawsetcolour(colour,action));
}


static os_error *displ_settruefontcol(font fonth, draw_coltyp foregrd,
                                           draw_coltyp backgrd)
{
 return(Draw_setcolour(fonth,backgrd,foregrd,14));
}



/* ------------------------------------------------------------------------- */

int flatness;
int minwidth;
int rotated;
int drawtypemask;


os_error *do_objpath(draw_objptr,transformstr*,processstr*);
os_error *do_objsprite(draw_objptr hdrptr,transformstr*,processstr*);
os_error *do_objtranssprite(draw_objptr hdrptr,transformstr*,processstr*);
os_error *do_objtext(draw_objptr hdrptr,transformstr*,processstr*);
os_error *do_objtranstext(draw_objptr hdrptr,transformstr*,processstr*);
os_error *do_objtransjpeg(draw_objptr hdrptr,transformstr *tr,processstr *pr);


/* Font catalogue - type defined in drawMenu */
fontcatstr fontcat;
BOOL fontcatseen;

/* Record font catalogue */
void note_fontcat(draw_objptr hdrptr)
{
  int ptr = sizeof(draw_fontliststrhdr);
  int fontNum;
  int size;

  if (hdrptr.bytep == 0)
  {
   fontcatseen = FALSE;
   memset(&fontcat,0,sizeof(fontcatstr));

   return;
  }

  size=hdrptr.fontlistp->size;
  fontcat.list_size = 0;
  while (ptr < size)
  {
    if ((fontNum = (int)hdrptr.bytep[ptr++]) == 0) break;
    fontcat.name[fontNum] = hdrptr.bytep + ptr;
    ptr += strlen(fontcat.name[fontNum]) + 1;
    if (fontNum > fontcat.list_size) fontcat.list_size = fontNum;
  }
  fontcat.menu_size = fontcat.list_size;
  fontcatseen = TRUE;
}

/* ------------------------------------------------------------------- */
/*                                                                     */
/* Draw the object 'hdrptr-> ' whose origin is (orgx,orgy)             */
/*                                                                     */
/*  The object is:                                                     */
/*    a PATH                                                           */
/*    a line of TEXT                                                   */
/*    a GROUPing of objects                                            */
/*                                                                     */
/*    (orgx,orgy) takes into acount window and scroll bar positions.   */
/*    (orgx,orgy) & clip(x0,y0,x1,y1) are in dBase coordinates.        */
/*                                                                     */
/* This version is modified from the one in c.drawDispl, to use a      */
/* different sort of error block, and to call the unknown object       */
/* handler if it has been defined. -> TRUE if no error                 */
/*                                                                     */
/* ------------------------------------------------------------------- */ 


BOOL  do_objects(draw_objptr hdrptr,draw_objptr limit,
                    transformstr * tr,processstr * pr,
                    int clip_x0,int clip_y0,int clip_x1,int clip_y1,
                    Draw_error *err)
{
 os_error * oserr=0;
 BOOL     ok     =TRUE;
 int      orgx;
 int      orgy;

 orgx=0;
 orgy=0;


 for(; ok && hdrptr.bytep < limit.bytep;
          hdrptr.bytep += hdrptr.objhdrp->size)
    { 


/*

 dprintf(4,"type=%d visible=%d",hdrptr.objhdrp->tag & drawtypemask,visible(hdrptr)); 


 dprintf(6,"size=%d",hdrptr.objhdrp->size); 


 dprintf(5,"x0=%x x1=%x y0=%x y1=%x",
                    hdrptr.objhdrp->bbox.x0,
                    hdrptr.objhdrp->bbox.x1,
                    hdrptr.objhdrp->bbox.y0,
                    hdrptr.objhdrp->bbox.y1);
*/


      switch(hdrptr.objhdrp->tag & drawtypemask)
      {
        case draw_OBJTEXT:
          if (visible(hdrptr)) oserr = do_objtext(hdrptr,tr,pr);
          break;

        case draw_OBJTRANSTEXT:
          if (visible(hdrptr)) oserr = do_objtranstext(hdrptr,tr,pr);
          break;

        case draw_OBJPATH:
          if(visible(hdrptr)) oserr = do_objpath(hdrptr,tr,pr);
          break;

        case draw_OBJSPRITE:
          if (visible(hdrptr)) oserr = do_objsprite(hdrptr,tr,pr);
          break;

        case draw_OBJTRANSSPRITE:
          if(visible(hdrptr)) oserr=do_objtranssprite(hdrptr,tr,pr);
          break;

        case draw_OBJTRANSJPEG:
          if(visible(hdrptr)) oserr=do_objtransjpeg(hdrptr,tr,pr);
          break;

        case draw_OBJGROUP:
          if (visible(hdrptr))
            { draw_objptr objptr,objlim;

              objptr.bytep = hdrptr.bytep + sizeof(draw_groustr);
              objlim.bytep = hdrptr.bytep + hdrptr.objhdrp->size;

              /* Render the objects in the group */
              ok = do_objects(objptr,objlim,tr,pr,
                              clip_x0,clip_y0, clip_x1,clip_y1, err);
            }
          break;

        case draw_OBJTAGG:
         /* if (visible(hdrptr)) */
            { draw_objptr objptr,objlim;


              objptr.bytep = hdrptr.bytep + sizeof(draw_taggstrhdr);
              objlim.bytep = objptr.bytep + objptr.objhdrp->size;

              /* Render the (one) object within the tagged object */
              ok = do_objects(objptr,objlim,tr,pr,
                              clip_x0,clip_y0, clip_x1,clip_y1, err);
            }

          break;

        case draw_OBJTEXTCOL:
        {
          draw_bboxtyp clipBox;
          clipBox.x0 = clip_x0;
          clipBox.y0 = clip_y0;
          clipBox.x1 = clip_x1;
          clipBox.y1 = clip_y1;

          if(visible(hdrptr)) oserr=do_objtextcol(hdrptr,tr,pr,&clipBox);
          break;
        }

        case draw_OBJTEXTAREA:
        {
          draw_bboxtyp clipBox;
          clipBox.x0 = clip_x0;
          clipBox.y0 = clip_y0;
          clipBox.x1 = clip_x1;
          clipBox.y1 = clip_y1;

          if(visible(hdrptr)) oserr=do_objtextarea(hdrptr,tr,pr,&clipBox);
          break;
        }

        case draw_OBJFONTLIST:
          note_fontcat(hdrptr);
          break;

        default: /* Call unknown object handler, if there is one */
          if (visible(hdrptr) && dr_unknown_handler != NULL)
            ok = dr_unknown_handler((void*)hdrptr.bytep, err);
          break;

      }

      if (oserr)
      {
        DrawFile_OSError(err, *oserr);
        ok = FALSE;
        break;
      }
    }
  return(ok);
}


/* ------------------------------------------------------------------- */ 


/* Fill and/or outline a path */


static void displ_unpackpathstyle(draw_objptr hdrptr, draw_jointspec *jspecp)
{
  int joincapwind    = hdrptr.pathp->pathstyle.joincapwind;

  jspecp->join       = (char)(joincapwind & packmask_join);
  jspecp->endcap     = (char)((joincapwind & packmask_endcap) >> packshft_endcap);
  jspecp->startcap   = (char)((joincapwind & packmask_startcap) >> packshft_startcap);
  jspecp->reserved   = 0;
  jspecp->mitrelimit = 0xA0000;  /* Mitre limit=10.0 (post script default) */
  jspecp->endtricapwid =
  jspecp->starttricapwid = (short)(hdrptr.pathp->pathstyle.tricapwid << 4);
  jspecp->endtricaphei =
  jspecp->starttricaphei = (short)(hdrptr.pathp->pathstyle.tricaphei << 4);
}



os_error *do_objpath(draw_objptr objhdr,transformstr * tr,processstr *pr)
{
 os_error * err;
 draw_coltyp fillcol = objhdr.pathp->fillcolour; /* BBGGRRxx / 0xFFFFFFFF */
 draw_coltyp linecol = objhdr.pathp->pathcolour;
 int linewid         = objhdr.pathp->pathwidth;
 int joincapwind     = objhdr.pathp->pathstyle.joincapwind;


 err=NULL;


 fillcol=processcolour(fillcol,pr);

 if(fillcol!=DXTRANSPARENT)
 {                                                         /* fill the path */
  err=displ_settruecol(0,fillcol);
  if(err) return(err);

  /* this is most questionable, flatness could be zero */

  err=Draw_fill((int*)address_pathstart(objhdr),tr,0x30 | ((joincapwind & packmask_windrule) >> 5),flatness,DRAWMOD_OSUNITS);
  if(err) return(err);
 }

 linecol=processcolour(linecol,pr);

 if(linecol!=DXTRANSPARENT)
 {                                                    /* stroke the outline */
  draw_jointspec jointspec;

  err=displ_settruecol(0,linecol);
  if(err) return(err);

  displ_unpackpathstyle(objhdr, &jointspec);

  if(linewid<minwidth && !(joincapwind & 0x3C)) linewid=0;
    
  err=Draw_stroke((int*)address_pathstart(objhdr),linewid,DRAWMOD_OSUNITS |0x38,(int*)&jointspec,tr,(int)address_dashstart(objhdr),flatness);

  if (err) return(err);
 }
 return(err);
}


/* ------------------------------------------------------------------- */ 


#ifdef NEVER

/* Set size & spacing of VDU5 characters, units are pixels */

os_error *displ_setVDU5charsize(int xsize,int ysize, int xspace,int yspace)
{ os_error *e;
  currentmode.gcharaltered = 1; /* so we restore before next pollwimp */

  e=NULL;

//  e = bbc_vduq(23,17,7,2,xsize,xsize>>8,ysize,ysize>>8,0,0);
//  if (!e) bbc_vduq(23,17,7,4,xspace,xspace>>8,yspace,yspace>>8,0,0);

  return(e);
}


#endif


/* Draw a line of text */

/* If the font can't be found (ie 'font disc not present' or font name  */
/* came from a fontlist object), use the system font. This seems kinder */
/* than aborting, or poping up an error box.                            */
/* Any other errors are passed back.                                    */

os_error *do_objtext(draw_objptr hdrptr,transformstr * tr,processstr *pr)
{
 os_error *    err;
 transformstr  trans;
 transformstr  result;
 coordstr      c;
 int           textcol;
 int           backgrd; /* a hint (kludge) to font munger */
 font          fonth=0;
 int           i;
 char        * p;


 err=NULL;


 textcol=hdrptr.textp->textcolour;
 backgrd=hdrptr.textp->background;

 textcol=processcolour(textcol,pr);
 if(textcol==DXTRANSPARENT) return(0); /* nothing to plot */
 backgrd=processcolour(backgrd,pr);


// if(fontcatseen && hdrptr.textp->textstyle.fontref)
// changed to force system font to go through this code as well as outline fonts

 {
  for(i=0;i<2;i++)
  {
   if(fontcatseen && hdrptr.textp->textstyle.fontref)
			{
    if(i<1) p=fontcat.name[hdrptr.textp->textstyle.fontref];
    else    p="Arial";
			}
			else
			{
				p="Arial";
				i=1;
			}


   if(rotated)
    err=Draw_find(p,xscaleFR(hdrptr.textp->fsizex),
                    yscaleFR(hdrptr.textp->fsizey),
                    0,0,
                    &fonth
                   );

   else
    err=Draw_find(p,xscaleFR(scale(hdrptr.textp->fsizex,tr->a,0x10000)),
                    yscaleFR(scale(hdrptr.textp->fsizey,tr->d,0x10000)),
                    0,0,
                    &fonth
                   );
   if(!err) break;
  }

  if(!err)
  { 
  /* err=font_setfont(fonth); */

   if(!err)
   {
    err=displ_settruefontcol(fonth,textcol,backgrd);
 
    if(!err)
    {
     if(rotated)
     {
      tr_setshift(&trans,hdrptr.textp->coord.x,hdrptr.textp->coord.y);
 
      tr_multiply(&result,&trans,tr);

      c.x=result.e;  /* convert to font from Draw units */
      c.y=result.f;

      result.e=0;
      result.f=0;

      err=Draw_painttrans(&hdrptr.textp->text[0],
                  font_OSCOORDS|
                  (((pr->fontbits & 0x1) && !pr->isprinter)?font_BACKBLEND:0),
                  c.x/256,c.y/256,(font_transform*)&result);
     }
     else
     {
      c.x=hdrptr.textp->coord.x;
      c.y=hdrptr.textp->coord.y;

      tr_transformset(&c,1,tr);

      err=Draw_paint(&hdrptr.textp->text[0],
                 font_OSCOORDS|(((pr->fontbits & 0x1) && !pr->isprinter)?
                                 font_BACKBLEND:0) /* | font_ABS */,
                                 c.x>>8,
                                 c.y>>8
                                );
     }
    }
   }
   Draw_lose(fonth);
   return(err);
  }
 }

#ifdef NEVER

/* either system font specified, or fancy font could not be found */
/* so render in the system font (ie scaled VDU5 characters)       */

  {
   int xsize;
   int ysize;

   xsize=scale(hdrptr.textp->fsizex,tr->a,0x10000);
   xsize/=currentmode.pixsizex;
   ysize=scale(hdrptr.textp->fsizey,tr->d,0x10000);
   ysize/=currentmode.pixsizey;

   /* xsize,ysize in pixels. pixsizex,pixsizey in dBase coords per pixel */
 
   err=displ_settruecol(0,textcol);
   if (err) return(err);

   /* assume char base line is row 7 (of 8) */

   c.x=hdrptr.textp->coord.x;
   c.y=hdrptr.textp->coord.y+7*(int)hdrptr.textp->fsizey/8;

   tr_transformset(&c,1,tr);

   err=bbc_move(c.x>>8,c.y>>8);
   if (err) return(err);
   err = displ_setVDU5charsize(xsize,ysize, xsize,ysize);
   if (err) return(err);
   err = bbc_stringprint(&hdrptr.textp->text[0]);
  }

#endif


 return(err);
}


#define DRAW_KERN      0x1
#define DRAW_RIGHTLEFT 0x2


os_error *do_objtranstext(draw_objptr hdrptr,transformstr * tr,processstr *pr)
{
 os_error   * err;
 int          textcol;
 int          backgrd;
 transformstr trans;
 transformstr result;
 font         fonth=0;
 int          x;
 int          y;
 int          fontbits;
 char       * p;
 int          i;

 err=NULL;

 textcol=hdrptr.transtextp->textcolour;
 backgrd=hdrptr.transtextp->background; /* a hint (kludge) to font munger */
 fontbits=hdrptr.transtextp->fontflags;


 textcol=processcolour(textcol,pr);
 if(textcol==DXTRANSPARENT) return(0); /* nothing to plot */
 backgrd=processcolour(backgrd,pr);


 if(fontcatseen && hdrptr.transtextp->textstyle.fontref)
 {
  for(i=0;i<2;i++)
  {
   if(i<1) p=fontcat.name[hdrptr.transtextp->textstyle.fontref];
   else    p="Arial";


   err=Draw_find(p,xscaleFR(hdrptr.transtextp->fsizex),
                   yscaleFR(hdrptr.transtextp->fsizey),
                   0,0,
                   &fonth
                 );

   if(!err)  break;
  }

  if(!err)
  { 
 /*  err=font_setfont(fonth); */

   if(!err)
   {
    err=displ_settruefontcol(fonth,textcol,backgrd);

    trans=*((transformstr*)&hdrptr.transtextp->trans);
    trans.e+=hdrptr.transtextp->coord.x;   /* are these in the same units? */
    trans.f+=hdrptr.transtextp->coord.y;

    tr_multiply(&result,&trans,tr);

    x=result.e;       /* doing this to get rid of multiple   */
    y=result.f;       /* cached font with trivial transforms */

    result.e=0;       /* convert to font from Draw units */
    result.f=0;

    err=Draw_painttrans(&hdrptr.transtextp->text[0],
               font_OSCOORDS |
               (((pr->fontbits & 0x1) && !pr->isprinter)?font_BACKBLEND:0) |
               ((fontbits & DRAW_KERN)?font_AUTOKERN:0) |
               ((fontbits & DRAW_RIGHTLEFT)?font_RIGHTLEFT:0),
                            x/256,
                            y/256,
                            (font_transform*)&result
                           );
   }

   Draw_lose(fonth);

   return(err);
  }
 }
 return(NULL);
}



#ifdef NEVER

/*********************************************
 * Call the SWI to build a translation table *
 *********************************************/

static os_error * selecttable(int sourcemode, int *sourcepal, int destmode,
                      int *destpal, char *transtab)
{
 os_regset regs ;

 regs.r[0] = sourcemode ;
 regs.r[1] = (int) sourcepal ;
 regs.r[2] = destmode ;
 regs.r[3] = (int) destpal ;
 regs.r[4] = (int) transtab ;
 regs.r[5] = 0;

 return (os_swix(0x40740/*ColourTrans_SelectTable*/, &regs)) ;
}


#endif



#ifdef NEVER

static os_error * setuppixtrans(sprite_header * sprite,int * pixtrans,
                                       processstr * pr,int ** pixtransp)
{
 os_error   * err;
 int          palentries;
 int          i;
 int       *  from;
 int       *  to;
 int          spbpp=0;
 int          spbitmask;


 err=NULL;

 *pixtransp=NULL;

 palentries=(sprite->image-sizeof(sprite_header))/8;
 if((palentries>256) || (palentries<0)) palentries=0;

 if(palentries==0)
 {
  getbpp(sprite->mode,&spbpp,&spbitmask);
 }

 if(pr->mode==PROCESSNONE && (palentries>0 || spbpp>3))
 {
  *pixtransp=NULL;
 }
 else
 {
  if(palentries==0)
  {
   /* must be spbpp<=3 */
   genpalbpp(pixtrans,spbpp,&palentries);
   *pixtransp=pixtrans;
//			cprintf("draw gen pal %d spbpp %d",palentries,spbpp);

   /* no palette, use default for sprite.mode */
  }
  else
  {
   getbpp(sprite->mode,&spbpp,&spbitmask);
  }

  if(pr->mode!=PROCESSNONE)
  {
//			cprintf("draw process");

   if(!*pixtransp)
   {
    from=(int*)(((char*)sprite)+sizeof(sprite_header));
    to=pixtrans;            

    for(i=palentries;i>0;i--)
    {
     *to++=*from;
     from+=2;
    }
    *pixtransp=pixtrans;
   }

   genpal256(pixtrans,ln2(spbpp),&palentries);

   for(i=0;i<palentries;i++) 
   {
    pixtrans[i]=processcolour(pixtrans[i] & 0xFFFFFF00,pr);
   }
  }
 }

 return(err);
}

#endif


/* plot a sprite */

os_error *do_objsprite(draw_objptr hdrptr,transformstr * tr,processstr *pr)
{
 os_error       * err;
 sprite_id        id;
 sprite_factors   factors;
 sprite_info      info;
// int              pixtrans[256];
// int            * pixtransp;
 transformstr     sptr;
 transformstr     result;
 int              dx;
 int              dy;
 int              claimed;


 id.tag   =sprite_id_addr;
 id.s.addr=&hdrptr.spritep->sprite;

 err=sprite_readsize((sprite_area*)0xFF, /* this op needs no spArea */
                           &id,                /* pass address of sprite  */
                           &info               /* result block            */
                          );

 factors.xmag=hdrptr.spritep->bbox.x1-hdrptr.spritep->bbox.x0;
 factors.xdiv=info.width;
 factors.ymag=hdrptr.spritep->bbox.y1-hdrptr.spritep->bbox.y0;
 factors.ydiv=info.height;

 getdeltas(info.mode,&dx,&dy);

 tr_setshift(&sptr,hdrptr.spritep->bbox.x0,hdrptr.spritep->bbox.y0);
 tr_addscale(&sptr,factors.xdiv*(dx<<8),factors.xmag,
                   factors.ydiv*(dy<<8),factors.ymag);

 tr_multiply(&result,&sptr,tr);


 if(!dr_enhanced_handler || 
    ((err=dr_enhanced_handler(&hdrptr,&result,pr,&claimed))==NULL && !claimed))
 {
#ifdef NEVER
  err=setuppixtrans(&hdrptr.spritep->sprite,pixtrans,pr,&pixtransp);
  if(err) return(err);

  if(rotated)
  {
   err=sprite_put_trans((sprite_area*)0xFF,&id,
                                            8|pr->spritebits,
                                            (sprite_transform*)&result,
                                            (sprite_pixtrans*)pixtransp);
  }
  else
  {
   /* there are problems with the factors being too big */


   factors.xmag=result.a;
   factors.xdiv=scale(0x10000,180/dx,cvdu.xdpi);
   factors.ymag=result.d;
   factors.ydiv=scale(0x10000,180/dy,cvdu.ydpi);
// cprintf("xdiv %x xdpi %d cdpi %d",factors.xdiv,180/dx,cvdu.xdpi);

 /*   c.x=hdrptr.spritep->bbox.x0;
    c.y=hdrptr.spritep->bbox.y0;
    tr_transformset(&c,1,tr);  */

 /*  c.x=result.e;
     c.y=result.f; */

   err=sprite_put_scaled((sprite_area*)0xFF, /* this op needs no spArea */
                           &id,              /* pass address of sprite  */
                           8|pr->spritebits, /* GcolAction=STORE thro mask */
                           result.e>>8, /* was c.x */
                           result.f>>8, /* was c.y */
                           &factors,
                           (sprite_pixtrans*)pixtransp
                         );
  }
#endif
 }

 return(err);
}




/* plot a sprite */

os_error *do_objtranssprite(draw_objptr hdrptr,transformstr *tr,processstr *pr)
{ 
 os_error       * err;
// sprite_id        id;
// transformstr     result;
// int              pixtrans[256]; /* pixel conversion tab                */
// int            * pixtransp;
 int              claimed;


 err=NULL;

 if(!dr_enhanced_handler || 
    ((err=dr_enhanced_handler(&hdrptr,tr,pr,&claimed))==NULL && !claimed))
 {
#ifdef NEVER
  err=setuppixtrans(&hdrptr.transspritep->sprite,pixtrans,pr,&pixtransp);
  if(err) return(err);

  id.tag=sprite_id_addr;
  id.s.addr=&hdrptr.transspritep->sprite;

  tr_multiply(&result,(transformstr*)&hdrptr.transspritep->trans,tr);

  err=sprite_put_trans((sprite_area*)0xFF,&id,8|pr->spritebits,
                (sprite_transform*)&result,(sprite_pixtrans*)pixtransp);
#endif
 }

 return(err);
}





/* plot a jpeg */


os_error *do_objtransjpeg(draw_objptr hdrptr,transformstr *tr,processstr *pr)
{ 
 os_error       * err=NULL;
// transformstr     result;
 int              claimed;
// jpeginfostr      info;

 if(!dr_enhanced_handler || 
    ((err=dr_enhanced_handler(&hdrptr,tr,pr,&claimed))==NULL && !claimed))
 {
#ifdef NEVER
  err=jpeg_info(hdrptr.transjpegp->data,hdrptr.transjpegp->len,&info);
  if(!err)
  {
   tr_multiply(&result,(transformstr*)&hdrptr.transjpegp->trans,tr);
   err=jpeg_render(hdrptr.transjpegp->data,hdrptr.transjpegp->len,&result,(pr->jpegbits & ~0x1),pr);
  }
#endif
 }

 return(err);
}





