
#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <stdarg.h>

#include "os.h"

#include "deb.h"
#include "etc.h"

#include "oserr.h"
#include "osmem.h"


// code for OS memory allocation


#define AREACHUNK (128*1024*1024)

static int pagemask;
static int handlebits;

#define MAXHANDLES 32


os_error * Task_ChangeMemoryAreaSize(int handle,void * oldtop,void * newtop,void ** newtopp)
{
 os_error * err;
 unsigned int change;
 void     * base;

 err=NULL;

 newtop=(void*)((((int)newtop)+pagemask)&(~pagemask));
 if(newtop>oldtop)
 {
  change=(int)newtop-(int)oldtop;
  err=XVirtualAlloc(oldtop,change,MEM_COMMIT,PAGE_READWRITE,&base);
  *newtopp=newtop;
 }
 else
 if(newtop<oldtop)
 {
  change=(int)oldtop-(int)newtop;
  err=XVirtualFree(newtop,change,MEM_DECOMMIT);
  *newtopp=newtop;
 }
 
 return(err);
 USE(handle);
}


os_error * Task_DeleteMemoryArea(int handle)
{
 os_error * err;

 handlebits&=~(1<<handle);
 err=XVirtualFree((void*)(handle*AREACHUNK),AREACHUNK,MEM_RELEASE);

 return(err);
}


static void Task_CheckPageSize(void)
{
 SYSTEM_INFO sinfo;

 if(pagemask==0)
 {
  XGetSystemInfo(&sinfo);
  pagemask=sinfo.dwPageSize-1;
  cprintf("pagemask=%d",pagemask);
 }
}

os_error * Task_CreateMemoryArea(int * handle,void ** base,unsigned int max,int bits,char * name)
{
 os_error * err;
 int        mhandle;
 void     * pagebase;

// cprintf("CreateMemoryArea");

 Task_CheckPageSize();

 for(mhandle=1;mhandle<MAXHANDLES;mhandle++)
 {
  if((handlebits & (1<<mhandle))==0) break;
 }

 err=XVirtualAlloc((void*)(mhandle*AREACHUNK),AREACHUNK,MEM_RESERVE,PAGE_NOACCESS,&pagebase);
 if(!err)
 {
  handlebits|=(1<<mhandle);
  *handle=mhandle;
  *base=(void*)(mhandle*AREACHUNK);
 }

 return(err);

 USE(name);
 USE(bits);
 USE(max);
}


os_error * Task_MemoryLimit(int mhandle,void** limit)
{
 *limit=(void*)((mhandle+1)*AREACHUNK);
 return(NULL);
}



os_error * Task_MaxVM(__int64 * maxvm,__int64 * maxphys,__int64 * maxpage)
{

#ifdef W64

 MEMORYSTATUSEX mem;

	mem.dwLength=sizeof(mem);

 GlobalMemoryStatusEx(&mem);

 if(maxvm)   *maxvm=mem.ullTotalVirtual;
 if(maxphys) *maxphys=mem.ullTotalPhys;
 if(maxpage) *maxpage=mem.ullTotalPageFile;

#else

 MEMORYSTATUS mem;

	mem.dwLength=sizeof(mem);

 GlobalMemoryStatus(&mem);

 if(maxvm)   *maxvm=mem.dwTotalVirtual;
 if(maxphys) *maxphys=mem.dwTotalPhys;
 if(maxpage) *maxpage=mem.dwTotalPageFile;


#endif

 return(NULL);
}


